/**
	backwardsearch.hpp
	Purpose: Implementation of the BWT-Backward-Search with Variants

	@author Thomas Büchler
*/
#include <string>
#include <vector>
#include <iostream>
#include <list>
#include <stack>
#include <tuple>
#include <climits>
#include <sdsl/int_vector.hpp>

#include "iupac_meth.hpp"
#include "../structs/searchinterval.hpp"
#include "../structs/interval_list.hpp"
#include "../structs/transform.hpp"


#ifndef BAS_H
#define BAS_H

typedef unsigned int ui;


/**
 * \brief searches a pattern in the string using the bwt
 * 
 * implementation of the bwt-backward-seach wich supplies jumps from # to #
 * Note sa-numbering starts with 0. '[' und ']' are inclusive. 
 * The interval [0,0] represents the iv with the first letter in it.
 * 
 * \param pattern 	the pattern, which should be searched for
 * \param fm_index 	fm_index of the string (only .C, .wavelet_tree.rank, char2comp used)
 * \param jia 			jump index array
 * \param jis				jump index string
 * \param trans			maps that map a number of a marker to its number in a other numbering (rank in jis <> rank in bwt <> pos in jis)
 * \param maxN			number of 'N' that are allowed to match
 * \param calc_kmi	indicator wether this method is called from kmi-calculator
 * \param i					position of the pattern where search starts (-1 if start with end and no interval_list is given)
 * \param intervals	interval_list with wich searching is continued (default goes with i = -1)
 */
template <class fmi_type, class jia_type>
interval_list back_search(std::string& pattern, fmi_type & fm_index, jia_type & jia, std::string & jis, transform_id & trans, int maxN = 5,
			  bool calc_kmi = false, int i = -1, interval_list intervals = interval_list()){
using namespace std;
using namespace sdsl;  
  if(i == -1){
    // set the initial sa-interval as the complete array
    i		= pattern.length();
    ui l 	= 0;
    ui r 	= fm_index.size()-1;
    searchinterval iv = searchinterval(searchinterval::bounds(l,r));
    
    // add initial sa-iv to list of sa
    intervals.add(iv);      
  }
  
  
  if(calc_kmi) maxN = pattern.length();//map every N in kmi calculation

  
  //iterate over the letters of the pattern
  while(i > 0 && !intervals.empty()){   
    //going to the next letter of the pattern
    i--;
    
    //cout << intervals.set.size() <<"\t";
    

    
    //list<searchinterval>	extra_intervals;	//container for intervals  with iupac letters that are not ATGC    
    interval_list	extra_intervals;
    
    //iterate over intervals and make search step
    for(list<searchinterval>::iterator iter = intervals.begin(); iter != intervals.end(); ++iter){
      searchinterval* iv 	= &(*iter);
      
      {//add intervals for IUPAC letters == handle_IUPACs<ji_typ>(i, pattern[i], ji, iv, extra_intervals, maxN); 
				string iupec_cs = getPossibleIUPAC_Letters(pattern[i]);	//matching iupac chars
				//values for SNP-interval
				ui l;
				ui r;
				for(unsigned int j = 0; j < iupec_cs.length(); j++){
								
					char 	iupec_c 	= iupec_cs[j]; // a matching IUPAC Letter
					
					int N 		= iv->N;
					if (iupec_c 	== 'N') N++;
					if (N 		> maxN) continue; //If to Many N matched DO nothing
					
					//calculate new interval
					ui 	C_iupec_c	= fm_index.C[ fm_index.char2comp[ iupec_c]];
					l 	= C_iupec_c 	+ fm_index.wavelet_tree.rank(iv->lr.first,	iupec_c);
					r 	= C_iupec_c 	+ fm_index.wavelet_tree.rank(iv->lr.second+1,iupec_c);
					
					
					if(l == r); // DO nothing
					
					else{
						//searchinterval snpiv  = searchinterval(searchinterval::bounds(l,r),iv->ret, N);
						searchinterval snpiv  = searchinterval(searchinterval::bounds(l,r-1), N);
						
						//Add SNP interval to extra_intervals	
						//extra_intervals.push_back(snpiv); 
						extra_intervals.add(snpiv); 	    
					}
				} 
      }
      
      ui C_p_i = fm_index.C[ fm_index.char2comp[ pattern[i]]];
      //calculate new interval      
      iv->lr.first	= C_p_i 	+ fm_index.wavelet_tree.rank(iv->lr.first,	pattern[i]);
      iv->lr.second 	= C_p_i 	+ fm_index.wavelet_tree.rank(iv->lr.second+1,pattern[i]) -1;
      
      //erasing interval if its empty
      if (iv->lr.first > iv->lr.second || iv->lr.second == UINT_MAX){								
				iter 		= intervals.erase(iter);
				iter--;
      }
    }        
    
    
    //intervals.splice(intervals.end(),extra_intervals); //merge lists
    intervals.merge(extra_intervals);    
    
    
    
    //iterate over singleintervals and make search step
    for(auto sp = intervals.begin_single(); sp != intervals.end_single(); ++sp){
      
      if(sp->pos < 0){	//delete interval
				sp 		= intervals.erase(sp);
				sp--;	    
				continue;
      }
      char c = jis[sp->pos];
      
      if( c != '#'){ //no jump		
				if(!matches(c, pattern[i])){
					sp 		= intervals.erase(sp);			//erase if it deos not match 
					sp--;
				}else{
					if(c == 'N' && sp->N >= maxN){
						sp 		= intervals.erase(sp);			//erase because too many N
						sp--;
					}else{
						sp->pos--;	//if match make backward step	
						if(c == 'N') sp->N++;
					}	  
				}	
      }else{	
				ui		rb_loc = trans.ps2rb.find(sp->pos)->second; 	//rank of jump-# in bwt
				bool	to_vari = jia.to_vari(rb_loc);
				
				if(to_vari){	//jump to alternative
					
					int_vector<0> targets = jia.targets(rb_loc);	//targets
					
					for(ui t = 0; t < targets.size(); t++){		//add each target
						ui rb_alt 	= targets[t];					//rank of jumptarget-# in bwt
						
						ui rs_alt 	= trans.rb2rs[rb_alt];				//rank of jumptarget-# in seq -1 (forwad matching needs the # one before)
						ui new_pos	= trans.rs2ps[rs_alt]-1;	    
							
						searchposition next = searchposition( new_pos, sp->N, sp->ret ); //define target pos
						
						if(!jia.to_vari(rb_alt))//jump from loco to loci (happens in sv-handling) -> no retpos
							next.ret.push_back(sp->pos -1);
						intervals.add( next );
					}	
						//jumps added, position can be deleted
					sp 		= intervals.erase(sp);
					sp--;
				}else{
					
					if(sp->ret.empty()){
					
						ui rb_all	= trans.ps2rb.find(sp->pos)->second;	 	//rank of allele marking-# in bwt
						
						int_vector<0> targets = jia.targets(rb_all);	//targets	  
						
						for(ui t = 0; t < targets.size(); t++){
						
							ui rb_targ 	= targets[t];		//rank of jumptarget-# in bwt  	  	    
							
							ui new_pos = trans.rb2ps[rb_targ] -1;		//jump
							
							searchposition next = searchposition( new_pos, sp->N, sp->ret); //define target iv
							
							intervals.add( next );
						}
						
						//jumps added, position can ve deleted
						sp 		= intervals.erase(sp);
						sp--;	    
					
					}else{
						ui retpos = sp->ret.back();
						sp->ret.pop_back();
						searchposition next = searchposition( retpos, sp->N, sp->ret);
						intervals.add( next );
						//jumps added, position can ve deleted
						sp 		= intervals.erase(sp);
						sp--;	    
					}
				}
			}	
		}
      
      
    
    if(i > 0 || calc_kmi) { //handle_Markers intervals;
			// C[#] stored because it is used often
			//int C_hash	= fm_index.C[ fm_index.char2comp['#']];
					
			//iterate over each interval, to check where jumps are needed
			for(auto iv = intervals.begin(); iv != intervals.end(); ++iv){
				
				//checking if a # occours in the iv
				ui firstHash	= fm_index.wavelet_tree.rank(iv->lr.first,'#');
				ui lastHash 	= fm_index.wavelet_tree.rank(iv->lr.second+1,'#');
				
				// Loop  over markers
				for(ui h = firstHash; h < lastHash; h++){
					// FUTURE FEATURE: My jump from intervals of # ???
					
					
					//look up the jump targets for each #
					int_vector<0> targets = jia.targets(h);
					
					
					if(jia.to_vari(h)){	
						//if jump goes to alternativ then return interval is current hash
						//searchinterval::bounds ret = searchinterval::bounds( C_hash + h, C_hash + h);
						//iv->ret.push_back(ret);
						
						ui ret_pos = trans.rb2ps[h]-1; 
						
						for(size_t k = 0; k < targets.size(); k++){	 
						
							//FUTURE FEATURE: May let targets be real intervals ???
							ui t 		= targets[k];
							
							//searchinterval target_iv 	= searchinterval(make_pair<int,int>(C_hash + t, C_hash +t),iv->ret);
							searchposition target_pos = searchposition(trans.rb2ps[t]-1, iv->N);
							target_pos.ret.push_back(ret_pos);
							
						
							//add jump interval to list of intervals		
							//intervals.add(target_iv);	
							//add(intervals,target_iv, ji, ps2rb,  rs2rb, rb2rs,  rb2ps, rs2ps);
							intervals.add(target_pos);
						} 
						//iv->ret.pop_back();
						//continue;
					}else{ 	//to_sequenz
					
						for(size_t k = 0; k < targets.size(); k++){	 	
							//FUTURE FEATURE: May let targets be real intervals ???
							ui t 		= targets[k];
							//searchinterval target_iv 	= searchinterval(make_pair<int,int>(C_hash + t, C_hash +t), iv->ret);    
							
						
							//add jump interval to list of intervals							    
							//intervals.push_back(target_iv);	  
							//intervals.add(target_iv);		
							//add(intervals, target_iv, ji, ps2rb,  rs2rb, rb2rs,  rb2ps, rs2ps);  
							searchposition target_pos = searchposition(trans.rb2ps[t]-1, iv->N);
							intervals.add(target_pos);
						} 
					}
				}
			}      
    }    
  }
  return intervals;
}

#endif
