/**
	constuct_jis.cpp
	Purpose: Construction of the jump-index-string

	@author Thomas Büchler
*/
#include "construct_jis.hpp"
#include "iupac_meth.hpp"
#include "../structs/vcfl.hpp"
#include "../structs/construct_jia.hpp"
#include "../structs/position_helper.hpp"

#include <vector>
#include <stdio.h> //remove
#include <sys/stat.h> //filetest
#include <utility>      // std::pair
#include <stdexcept>      // std::invalid_argument

#include <limits.h> //UINT_MAX

#include <sdsl/int_vector.hpp>
#include "../programs/filenames.hpp"

using namespace std;
using namespace sdsl;

#define DUMMY_VALUE "DUMMY652956" //chromosome name for empty variation 
/**
 * \brief representation of a SNP * 
 */
struct snp{
	string chrom;
	unsigned int pos;
	char ref;
	char iupac;
	
	snp(){pos = UINT_MAX; chrom = DUMMY_VALUE;}
	
	snp(string vcf_line){
		vcfl v = vcfl(vcf_line);
		chrom = move(v.chrom);
		pos = v.pos;
		ref = v.ref[0];
		iupac = getSNPLetter(v.alt,ref);
	}
};

/**
 * \brief representation of a INDEL
 */
struct indel{
	string chrom;
	unsigned int pos;
	string id;
	string ref;
	deque<string> alt;
	
	indel(){pos = UINT_MAX;chrom = DUMMY_VALUE;}
	
	indel(string vcf_line){
		vcfl v = vcfl(vcf_line);
		v.genaltQ();
		
		chrom = move(v.chrom);
		pos = v.pos;
		id 	= move(v.id);
		ref = move(v.ref);
		alt = v.genaltQ();
	};
};

/**
 * \brief representation of a CNV
 */
struct cnv{
	string chrom;
	string id;
	unsigned int pos;
	unsigned int end;
	string ref;
	unsigned int max_copy;
	bit_vector copies;
	
	cnv(){pos = UINT_MAX;chrom = DUMMY_VALUE;}
	
	cnv(string vcf_line){
		vcfl v 		= vcfl(vcf_line);
		
		max_copy 	= v.maxCN();
		
		end 			= v.sv_end();	
		pos 			= v.pos;
		
		id				= move(v.id);
		chrom 		= move(v.chrom);
		ref 			= move(v.ref);
		
		copies = bit_vector(max_copy+1,0);
		
		copies[1] = true; //this is ref
		
		//alts has the form <CNx1>,<CNx2>,... with integers x1,x2..
		//search for N and > and read number between
		// set copies[xi] to true;
		size_t x_pos2 = 0;
	
		for(int i = 0; i < v.altCount(); i++){
			size_t x_pos1 = v.alt.find('N',x_pos2)+1;
			x_pos2 = v.alt.find('>',x_pos1);
			int x = stoi(v.alt.substr(x_pos1,x_pos2-x_pos1));
			copies[x] = true;			
		}	
	};
};

/**
 * \brief reads the next line in the snp file and returns a snp representation
 */
snp read_next_snp(sequence_streams & seq_s){
	string line;
	
	while(getline(seq_s.snp_in, line)){
    if(line[0] == '#') //ignore comment
      continue;		
		if(line!="")
			return(snp(line)); //return snp
  }
  return snp();
}

/**
 * \brief reads the next line in the indel file and returns a indel representation
 */
indel read_next_indel(sequence_streams & seq_s){
	string line;
	
	while(getline(seq_s.ind_in, line)){  		
    if(line[0] == '#') //ignore comment
      continue;			
		if(line!="")
			return(indel(line)); //return snp
  }
  return indel();
}

/**
 * \brief reads the next line in the sv file and returns a cnv representation
 */
cnv read_next_cnv(sequence_streams & seq_s){
	string line;
	
	while(getline(seq_s.sv_in, line)){  	
		if(line[0] == '#') //ignore comment
      continue;			
		if(line!="")
			return(cnv(line)); //return snp
  }
  return cnv();
}

/**
 * \brief decide wich variation is the next one
 * 
 * gets a representation of each variation type
 * and decides wich has the smalles position
 * 
 *\return a pair of the next variation and it's position
 */
pair<short,unsigned int> chooseNext(snp & snp, indel & ind, cnv & cnv, const string & fa_chrom){
	//no variations left
	if((snp.pos == UINT_MAX)&&
		 (ind.pos == UINT_MAX)&&
		 (cnv.pos == UINT_MAX))
		return make_pair(0,0); //no one 
		
	bool snp_in_next_chrom = snp.chrom != fa_chrom;
	bool ind_in_next_chrom = ind.chrom != fa_chrom;
	bool cnv_in_next_chrom = cnv.chrom != fa_chrom;
	
	if(snp_in_next_chrom && ind_in_next_chrom && cnv_in_next_chrom)
		return make_pair(0,0); //no one 
		
	
	if( (snp.pos < ind.pos || ind_in_next_chrom)&&
		  (snp.pos < cnv.pos || cnv_in_next_chrom)&&
			!snp_in_next_chrom)
		return make_pair(1,snp.pos); //SNP is next
	
		
	if( (ind.pos < cnv.pos || cnv_in_next_chrom) && 
		   !ind_in_next_chrom)
		return make_pair(2,ind.pos); //INDEL is next		
	
	if(!cnv_in_next_chrom)
		return make_pair(3,cnv.pos);	//CNV wins	
	else
		throw invalid_argument("SOME SHIT HAPPEND");
}

/**
 * \brief writes the IUPAC letter in jis
 */
void applySnp(sequence_streams & seq_s, snp & snp){
	
	if(!seq_s.check_ref(snp.iupac))
		throw invalid_argument("FAILURE\tREF DOES NOT MATCH\tSNP IGNORED\tchrom="  +seq_s.fa_chrom+ "\tpos="+to_string(snp.pos));   
	
	seq_s.write_char(snp.iupac); //write IUPAC
	seq_s.check_lay1_ended();
}

/** 
 * \brief writes a marker in jis and the variation in variation file 
 * writes the connection of the markers to the construct_jia struct
 */
void applyIndel(sequence_streams & seq_s, indel & indel, construct_jia & conjia, construct_position_helper & conph){
	
	if(!seq_s.check_ref(indel.ref))
		throw invalid_argument("FAILURE\tREF DOES NOT MATCH\tINDEL IGNORED\tchrom="  +seq_s.fa_chrom+ "\tpos="+to_string(indel.pos));   

	if(seq_s.write_in_jis){
		conjia.add_indel(false, seq_s.jis_markers, seq_s.lay2_markers-1, indel.alt.size() +1);
		conph.fa_pos_jis.push_back(seq_s.current_fa_pos-indel.ref.size());
		conph.vcf_id_jis.push_back(indel.id);
	}
	else {
		conjia.add_indel(true, seq_s.lay1_markers, seq_s.lay2_markers-1, indel.alt.size() +1);	
		conph.fa_pos_tmp1.push_back(seq_s.current_fa_pos-indel.ref.size());	
		conph.vcf_id_tmp1.push_back(indel.id);
	}
	
	seq_s.write_marker();
	
	for(unsigned int i = 0; i <= indel.alt.size(); i++)
		conph.fa_pos_tmp2.push_back(seq_s.current_fa_pos);
	
	seq_s.write_alternative(indel.ref);
	for(unsigned int i = 0; i < indel.alt.size(); i++)
		seq_s.write_alternative(indel.alt[i]);		
}


/** 
 * \brief writes markers in jis and switch the output stream from jis to tmp_lay1
 * 
 * writes the connection of the markers to the construct_jia struct
 */
void applyCnv(sequence_streams & seq_s, cnv & cnv, construct_jia & conjia, construct_position_helper & conph){
	
	conjia.add_cnv(seq_s.jis_markers , cnv.copies, seq_s.lay1_markers-1);
	
	for(unsigned int i = 1; i <= cnv.max_copy; i++){
		seq_s.write_marker();
		conph.vcf_id_jis.push_back(cnv.id);
		conph.fa_pos_jis.push_back(seq_s.current_fa_pos);
	}
	
	seq_s.write_in_jis = false;	
	seq_s.write_in_lay1_end = cnv.end;
}


/**
 * \brief construct the index string, writes it in file and return the index array
 * 
 * open streams to the fa and the 3 variation files
 * 
 * read a variation of each kind
 * decide wich one is handled first
 * process this variation
 * read the next variatio of this kind in
 * repeat this till all veriations are handled
 *
 * 
 */ 
jumpindexarray construct_jis(sequence_streams& seq_s, construct_streams & con_s){
	
	construct_jia conjia = construct_jia();
	construct_position_helper conph = construct_position_helper();

	seq_s.open_for_jis();
	conph.chromosome_names.push_back(seq_s.fa_chrom);
	con_s.log << "CROM: '" <<  seq_s.fa_chrom << "'" << endl;
	  
	snp next_snp = read_next_snp(seq_s);
	cnv next_cnv = read_next_cnv(seq_s);
	indel next_ind = read_next_indel(seq_s);
	
	unsigned int varcount = 0;
	while(true){
		varcount++;
		if(varcount % 100000 == 0)	
			cout << "\r" << varcount << " variations processed" << flush;  
		
		if(		//no chromosom matches
					seq_s.fa_chrom != next_snp.chrom 
			&& 	seq_s.fa_chrom != next_cnv.chrom 
			&& 	seq_s.fa_chrom != next_ind.chrom ){
			
			
			seq_s.copy_till_next_chrom();
		
			if(seq_s.write_in_lay1_ended){ //add end marker of SV
				seq_s.write_in_lay1_ended = false;
				conjia.cnv_set_rigth_end(seq_s.lay1_markers-1);
				conph.fa_pos_tmp1.push_back(seq_s.write_in_lay1_end);
				conph.vcf_id_tmp1.push_back("");
			}
			
			//do not read over chromosom end
			conph.seq_seperator.push_back(seq_s.jis_markers);
			seq_s.write_marker();
			conph.fa_pos_jis.push_back(seq_s.last_chrom_length);
			conph.vcf_id_jis.push_back("");
			
			if(seq_s.fa_chrom == "END_OF_FILE"){
				cout << "\r                                            \r"<< flush;
				break;
			}else{
				conph.chromosome_names.push_back(seq_s.fa_chrom);
				con_s.log << "CROM: '" <<  seq_s.fa_chrom << "'" << endl;				
			}
			continue;
		}
		pair<short, unsigned int> next = chooseNext(next_snp,next_ind,next_cnv,seq_s.fa_chrom);
		seq_s.copy_till_pos(next.second);
		
		if(seq_s.write_in_lay1_ended){ //add end marker of SV
			seq_s.write_in_lay1_ended = false;
			conjia.cnv_set_rigth_end(seq_s.lay1_markers-1);
			conph.fa_pos_tmp1.push_back(seq_s.write_in_lay1_end);
			conph.vcf_id_tmp1.push_back("");
		}
		//apply variation
		try{
			switch (next.first){
				case 1: //SNP
					applySnp(seq_s,next_snp);
					break;
				case 2: //INDEL 
					applyIndel(seq_s, next_ind, conjia, conph);
					break;
				case 3: //CNV
					applyCnv(seq_s, next_cnv, conjia, conph);
					break;
				case 0: //no one
					;
			}
		}catch(const invalid_argument & ia){
			con_s.log_error(ia.what());
		}
		//go to next variation
		switch (next.first){
			case 1: //SNP
				next_snp = read_next_snp(seq_s);
				break;
			case 2: //INDEL 
				next_ind = read_next_indel(seq_s);
				break;
			case 3: //CNV
				next_cnv = read_next_cnv(seq_s);
				break;
			case 0: //no one
				;
		}
				
	}
	
	//close streams
	seq_s.close();		
	//append sequence of tmp files
	seq_s.jis_out.open(seq_s.jis_filename,ios_base::out|ios_base::app);
	seq_s.tmp_lay1_in.open(seq_s.tmp_lay1_filename,ios_base::in);
  seq_s.jis_out << seq_s.tmp_lay1_in.rdbuf();
  seq_s.close();	
	
	
	seq_s.jis_out.open(seq_s.jis_filename,ios_base::out|ios_base::app);
	seq_s.tmp_lay2_in.open(seq_s.tmp_lay2_filename,ios_base::in);	
  seq_s.jis_out << seq_s.tmp_lay2_in.rdbuf();	
  seq_s.close();	
	
	//delete tmp files
  remove(seq_s.tmp_lay1_filename);	
  remove(seq_s.tmp_lay2_filename);
	
	conph.seq_seperator.push_back(seq_s.jis_markers+seq_s.lay1_markers-1);
	position_helper ph = conph.convert();
	string phs = con_s.out_dirname + POSITION_FNAME;
	if(!sdsl::store_to_file(move(ph),phs))
		con_s.log 		<< phs << " WAS NOT STORED CORRECTLY" << endl;	
		
		
	//return jia
	unsigned int markercount = seq_s.jis_markers+seq_s.lay1_markers+seq_s.lay2_markers;
	jumpindexarray jia = conjia.jia(markercount,seq_s.jis_markers,seq_s.lay1_markers);
	return move(jia);	  
}
