/**
	constuct_methods.hpp
	Purpose: Help methods that are used during construction of the index

	@author Thomas Büchler
*/
#include <vector>
#include <string>
#include <iostream>
#include <deque>
#include <map>
#include <sdsl/csa_wt.hpp>
#include "../structs/jumpindexarray.hpp"

#ifndef COM_H
#define COM_H

/**
 * \brief caclulates how many bits are needed to represent x
 */
int bitsneeded (unsigned int x);


/**
 * \brief genereates a map that maps pos_jis(#) -> rank_bwt(#)
 * 
 * makes a complete backward search trough the string
 * if an # dectected
 * the corresponding rank of the # can be retrieved from the fm_index
 * 
 * \param fm_index of the jis file
 * 
 */
template <class fm_typ>
std::map<int,int> pos2rankMapByStringScan(const fm_typ& fm_index ){
using namespace sdsl;
using namespace std; 
  //rank of the next # in the sequence
  int seq_pos = fm_index.size() -1; 
  //line in BWT (last # is rigth before EOF and so its in the first line)
  int line = 0; 
  
  map<int,int> pos2rankMap;    
  
  while(seq_pos >= 0){  
    char c = fm_index.wavelet_tree[line];  
    int c_rank =  fm_index.wavelet_tree.rank( line ,c);   
    
    if(c == '#')    
      pos2rankMap.insert(pair<int,int>(seq_pos-1,c_rank));
      
    line = fm_index.C[ fm_index.char2comp[c] ] + c_rank;   
    seq_pos--;
  }
  
  return move(pos2rankMap); 
};

/**
 * \brief genereates a map that maps pos_jis(#) -> rank_bwt(#)
 * 
 * fist calc string position of each # in the sequence with C and SA
 * hash_seq_pos = SA[ C[#] + rank(#,bwt) ] + 1
 * 
 * store pair of hash_seq_pos and rank(#,bwt) in a map
 * 
 */
template <class fm_typ>
std::map<int,int> pos2rankMapByUsingSA(const fm_typ& fm_index){  
using namespace std;
     map<int,int> pos2rankMap;     
     int c_pos = fm_index.C[ fm_index.char2comp['#']];
     
     int numberOfHash = fm_index.C[ fm_index.char2comp['#']+1]- c_pos;
     
     for(int  hash_bwt_rank = 0; hash_bwt_rank < numberOfHash; hash_bwt_rank++){
       int hash_seq_pos = fm_index[c_pos+hash_bwt_rank]; 			
       pos2rankMap.insert(pair<int,int>(hash_seq_pos,hash_bwt_rank));
      }
      return move(pos2rankMap);
};


/**
 * \brief generates a pos2rankMap
 * 
 * if use_sa is true pos2rankMap is generated with pos2rankMapByUsingSA()
 * else the string will be scanned for # with backward search 
 * 
 * \param fm_index of the jis file
 * \param use_sa sa is allowed to use
 * 
 */
template <class fm_typ>
std::map<int,int> pos2rankMap(const fm_typ& fm_index,const bool use_sa){
using namespace std;
  if(use_sa)
    return move(pos2rankMapByUsingSA(fm_index));
  else
    return move(pos2rankMapByStringScan(fm_index));
};


/**
 *\brief genereates the rank table
 * 
 * makes a complete backward search trough the string
 * if an # dectected, its the next smaller hash in the jis rank
 * the corresponding jis can be retrieved from the fm_index
 * 
 * \param numberOfHash in the jis file
 * \param fm_index of the jis file
 */
template <class fm_typ>
sdsl::int_vector<0> rankSeq2BwtByStringScan(const int & numberOfHash, const fm_typ& fm_index ){
using namespace std;
using namespace sdsl;
  int_vector<0> rankSeq2Bwt = int_vector<0>(numberOfHash,0,bitsneeded(numberOfHash));
  
  //rank of the next # in the sequence
  int seq_rank = numberOfHash-1; 
  //line in BWT (last # is rigth before EOF and so its in the first line)
  int line = 0; 
  
  while(seq_rank >= 0){  
    char c = fm_index.wavelet_tree[line];  
    int c_rank =  fm_index.wavelet_tree.rank( line ,c);   
    
    if(c == '#'){      
      rankSeq2Bwt[seq_rank] = c_rank;
      seq_rank--;
      
    }    
    line = fm_index.C[ fm_index.char2comp[c] ] + c_rank;    
  }
  
  return move(rankSeq2Bwt);  
};


/**
 *\brief genereates a table that maps a jis rank to a bwt rank (O(|jis|))
 * 
 * 
 * run over map elements (sorted by hash_seq_pos)
 * you get the rank(#,bwt) in order of rank(#,sequence)
 * 
 * \param numberOfHash in the jis file
 * \param fm_index of the jis file
 
template <class fm_typ>
sdsl::int_vector<0> rankSeq2BwtByUsingSA(const int & numberOfHash, const fm_typ& fm_index){
using namespace std;
  map<int,int> pos2rankMap = pos2rankMapByUsingSA(fm_index);
  return move(rankSeq2BwtByMap(numberOfHash,pos2rankMap));    
};*/


/**
 * \brief constructs the rankTable 
 * 
 * rankTable[rank(#,jis)]  = rank(#,bwt)  (the first # has rank 0)
 * 
 * if use_sa is true pos2rankMap is generated and the rankTable will be retreved from  the maps
 * else the string will be scanned for # with backward search 
 * 
 * \param numberOfHash in the jis file
 * \param fm_index of the jis file
 * \param use_sa sa is allowed to use
 * 
 
template <class fm_typ>
sdsl::int_vector<0> rankSeq2Bwt(const int & numberOfHash, const fm_typ& fm_index, const bool use_sa){
using namespace std;
  if(!use_sa)
    return move(rankSeq2BwtByStringScan(numberOfHash,fm_index));
   else
     return move(rankSeq2BwtByUsingSA(numberOfHash,fm_index)); 
}*/



sdsl::int_vector<0> invertVector(const sdsl::int_vector<0> & vector);



/** \brief help method to debug prints a vector*/
template <typename T>
void printVector(const std::vector<T> & v, const std::string & name){
  using namespace std;
  cout << "vector " << name <<" :<";
  for( unsigned int i = 0; i < v.size(); ++i ){	  
              cout << v[i];
	      if(i < v.size()-1)
		cout << ",";
  }
  cout << ">\n";
};

/** \brief help method to debug prints a deque*/
template <typename T>
void printQ(const std::deque<T> & v, const std::string & name){
  using namespace std;
  cout << "deque" << name <<":<";
  for( unsigned int i = 0; i < v.size(); ++i ){	  
              cout << v[i];
	      if(i < v.size()-1)
		cout << ",";
  }
  cout << ">\n";
};

/** \brief help method to debug prints a map*/
template <typename T,typename E>
void printMap(const std::map<T,E> & m, const std::string & name){  
  using namespace std;
  cout << "map" << name <<":<" << endl;
  for (auto i = m.begin();  i != m.end(); i++)
    cout<< i->first << " " << i->second << endl;
  
  cout << ">" << endl;
};

/** \brief prints all entries of all intervals on the std::cout*/
template <typename T> 
void printList(std::list<T>& list , const std::string & name){
  using namespace std;
  cout << "list " << name <<":<" << "\n" ;
  
 auto iter = list.begin();
 while(iter != list.end())
      cout << (*iter++);
 
  cout << ">" << endl;  
}

#endif

