/**
	forwardsearch.hpp
	Purpose: Implementation of the String Forwardsearch with variants

	@author Thomas Büchler
*/
#include <sdsl/suffix_arrays.hpp>
#include <sdsl/wt_huff.hpp>
#include <sdsl/io.hpp>

#include <iostream>
#include <fstream>
#include <iomanip> 
#include <sstream>  
#include <string>
#include <stack>
#include <list>    

#include "../structs/searchinterval.hpp"
#include "../structs/interval_list.hpp"
#include "../structs/searchresult.hpp"
#include "../structs/matchresult.hpp"
#include "../structs/jumpindexarray.hpp"
#include "../structs/compressedjia_select.hpp"
#include "../structs/compressedjia_rank.hpp"
#include "../structs/transform.hpp"
#include "../structs/position_helper.hpp"

#include "../methods/iupac_meth.hpp"

#ifndef FOS_H
#define FOS_H

typedef sdsl::csa_wt<sdsl::wt_huff<sdsl::bit_vector,sdsl::bit_vector::rank_1_type,sdsl::select_support_scan<1>,sdsl::select_support_scan<0>>,32,1024*1024>		fm_typ_sa_noisa;

/**
 * \brief Search position for forward search 
 */
struct forw_pos{
	unsigned int 	pos;
	std::deque<unsigned int> 	retpos = std::deque<unsigned int>();
	std::string 	hist;  
	forw_pos(unsigned int pos, std::deque<unsigned int> retpos, std::string hist):pos(pos),retpos(retpos),hist(hist){}
};

/**
 * \brief calculates the pos in the fasta file that relates to posistion pos in jis
 * 
 * \param pos jis position
 * \param rs_next_marker rank of the next marker
 * \param off jis offset to next marker
 * 
 * \return string for the output file  
 */
std::string start_in_jis_print(unsigned int pos, unsigned int rs_next_marker, unsigned int off, position_helper & pos_h);

/**
 * \brief calculates the chromosme name of the chromose where marker with the given rank is from
 * \param marker_rank jis rank of marker
 */
std::string chrom_name(unsigned int marker_rank, position_helper & pos_h);


/**
 * \brief calculates the site marker of the cnv that includes the marker with rank marker_rank
 */
template <class jia_type>
unsigned int cnv_site(unsigned int marker_rank, jia_type & jia){
	while(true){
		if(!jia.to_vari(marker_rank))//end of cnv
			return jia.targets(marker_rank)[0]; //first occ of cnv
		marker_rank++;
	}
}

/**
 * \brief calculates the pos in the fasta file that relates to posistion pos in jis 
 * 
 * \param pos jis position
 * \param rs_next_marker rank of the next marker
 * \param off jis offset to next marker
 * 
 * \return string for the output file  
 */
template <class jia_type>
std::string start_in_cnv_print(const unsigned int pos, unsigned int rs_next_marker, unsigned int off, jia_type & jia, position_helper & pos_h){
	unsigned int fa_pos = pos_h.fa_pos[rs_next_marker];
	unsigned int cnv_site_marker = cnv_site<jia_type>(rs_next_marker,jia);
	std::stringstream out;
	out << chrom_name(cnv_site_marker,pos_h) 
			<< '\t' << fa_pos - off 
			<< '\t' << pos+1  
			<< '\t' << "[starts in copy of "<< pos_h.vcf_id[cnv_site_marker] <<"]";
	return move(out.str());
};

/**
 * \brief calculates the pos in the fasta file that relates to posistion pos in jis
 * 
 * \param pos jis position
 * \param rs_next_marker rank of the next marker
 * \param off jis offset to next marker
 * 
 * \return string for the output file 
 */
template <class jia_type>
std::string start_in_var_print(const unsigned int pos, unsigned int rs_next_marker, unsigned int off, jia_type & jia, position_helper & pos_h){	
	using namespace std;
	std::stringstream out;
	
	//lookup site marker of the variation
	unsigned int rs_site = jia.targets(rs_next_marker)[0];
	//determine in wich variation of the site marker pos lies
	sdsl::int_vector<0> targets = jia.targets(rs_site);
	std::string alt = ":*?*";
	for(unsigned int i = 0; i < targets.size(); i++){
		if(targets[i] + 1 == rs_next_marker ){
			alt = i==0?":ref":":alt"+to_string(i);
		}
	}
	
	//if variation is inside a CNV 
	bool start_in_cnv = rs_site > pos_h.seq_seperator[pos_h.seq_seperator.size()-2];	

	if(!start_in_cnv) {
		out << chrom_name(rs_site, pos_h) 
				<< '\t' << pos_h.fa_pos[rs_next_marker] 
				<< '\t' << pos+1  
				<< '\t' << "["<< off << "-prefix in " << pos_h.vcf_id[rs_site] << alt << "]";
	}else{
		//determine site marker of the CNV
		unsigned int rs_cnv_site = cnv_site<jia_type>(rs_site,jia);		
		out << chrom_name(rs_cnv_site, pos_h) 
				<< '\t' << pos_h.fa_pos[rs_next_marker] 
				<< '\t' << pos+1  
				<< '\t'	<< "["<< off << "-prefix in " << pos_h.vcf_id[rs_site] << alt << " inside a copy of "<< pos_h.vcf_id[rs_cnv_site] <<"]";
	}
	return move(out.str());
}

/**
 * \brief calculates the pos in the fasta file that relates to posistion pos in jis
 * 
 * 
 * \param pos position in start_in_jis
 * \param jia jump index array (inverted for forward search, permutated for string search)
 * \param trans transformer (next_marker,rs2ps)
 * \param pos_h position_helper to get the ralting position in fasta file
 * 
 * \return string for the output file 
 */
template <class jia_type>
std::string get_ref_string_pos(const unsigned int pos, jia_type & jia, transform_id & trans, position_helper & pos_h){
	using namespace std;
	using namespace sdsl;
	//search for next marker.
	unsigned int rs_next_marker = trans.next_marker(pos);
	unsigned int off = trans.rs2ps[rs_next_marker] - pos;
	
	bool start_in_jis = rs_next_marker <= pos_h.seq_seperator[pos_h.seq_seperator.size()-2];
	bool start_in_var = rs_next_marker >  pos_h.seq_seperator[pos_h.seq_seperator.size()-1];
	bool start_in_cnv = rs_next_marker >  pos_h.seq_seperator[pos_h.seq_seperator.size()-2];
	
	if(start_in_jis)
		return start_in_jis_print(pos, rs_next_marker, off, pos_h);
	else if(start_in_var)
		return start_in_var_print<jia_type>(pos, rs_next_marker, off, jia, pos_h);
	else if(start_in_cnv)
		return start_in_cnv_print<jia_type>(pos, rs_next_marker, off, jia, pos_h);
	else
		return ("ERROR["+ to_string(pos)+ " starts nowhere get_ref_string_pos]\n"); //This should never be the case
};
	
/**
 * \brief exact forward search in jump index string
 * 
 * gets stringposition where matching should start and matches forward
 * 
 * \return string to write in file
 * 
 * \param pos		position where matching starts
 * \param pattern	matched pattern
 * \param jis		jump index string
 * \param jia		jump index array (inverted for forward search, permutated for string search)
 * \param trans	transformer (ps2rs,rs2ps)
 * \param pos_h position_helper to get the ralting position in fasta file
 */
template <class jia_type>
int forward_search(std::ofstream & out, unsigned int pos, const std::string & pattern, const std::string & jis, jia_type & jia, transform_id & trans, position_helper & pos_h){
	using namespace std;
	using namespace sdsl;
  //stringstream out;
	string fa_pos = get_ref_string_pos<jia_type>(pos,jia,trans,pos_h);
  
 
  
  forw_pos fp(pos,deque<unsigned int>(), fa_pos);	// initial interval    
  list<forw_pos> fpl;		//iv list  
  fpl.push_back(fp);
  
  //search:  
  for(unsigned int i = 0; i < pattern.length(); i++){	//max offset -length search steps
    for(list<forw_pos>::iterator fp= fpl.begin(); fp != fpl.end(); fp++ ){	//iterate over intervals
        
			//if no marker then check if match
			if( jis[fp->pos] != '#'){					
				if(!matches(jis[fp->pos], pattern[i])){	
					//erase if it does not match 
					fp 		= fpl.erase(fp);			
					fp--;
				}else{
					//go to next position
					fp->pos++;	
				}	  
			}else{
				int rs_loc = trans.ps2rs(fp->pos);
        bool	to_vari = jia.to_var[rs_loc];	
        
        if(to_vari){	//jump to alternative
					int_vector<0> targets = jia.targets(rs_loc);	//targets
					 
					for(unsigned int t = 0; t < targets.size(); t++){		//add each target
						unsigned int rs_t =  targets[t];
						bool to_cnv = rs_t < pos_h.seq_seperator[pos_h.seq_seperator.size()-1];
						unsigned int next_pos	= trans.rs2ps[rs_t] +1; //target = marker, +1 ist next position
						stringstream ss;
						ss <<  fp->hist <<  "[" << pos_h.vcf_id[rs_loc] << (to_cnv?
((rs_t < pos_h.seq_seperator[pos_h.seq_seperator.size()-2])?":skip copy":":read copy"):(t==0?":ref":(":alt" + to_string(t)))) << "]";
						forw_pos next = forw_pos( next_pos, fp->retpos, move(ss.str())); //define target iv
							
						if(!jia.to_vari(rs_t))//jump from loci to loci (happens in sv-handling) -> no retpos
							next.retpos.push_back(fp->pos +1);
						fpl.push_back( next );
					}	
					
						//jumps added, position can ve deleted
            fp 		= fpl.erase(fp);
            fp--;
        
        
        }else if( fp->retpos.empty()){	//first jump to locus
					//rank of jumptarget-# in bwt
					int rs_alt = trans.ps2rs(fp->pos);
          int_vector<0>	targets	= jia.targets(rs_alt);						
            
					
					for(unsigned int t = 0; t < targets.size(); t++){						
						int rs_alt =  targets[t];
						int next_pos	= trans.rs2ps[rs_alt] +1; //jump
						forw_pos next = forw_pos( next_pos, fp->retpos, fp->hist); //define target iv
						
						fpl.push_back( next );
					}
					
					//jumps added, position can ve deleted
					fp 		= fpl.erase(fp);
					fp--;	    
        
        }else{	//jumpback
					fp->pos	= fp->retpos.back();
					fp->retpos.pop_back();
					fp--;	//repeat iv	  	    
					}	
        }	
    }
  }
  
  
  //output:
  if(fpl.empty())
    out << "\tERROR: read cold not be found." << endl;
  else{
    for(list<forw_pos>::iterator fp= fpl.begin(); fp != fpl.end(); fp++ ){
      out << '\t'<< fp->hist << endl;
    }
  } 
  return fpl.size();
}

#endif
