/**
	iupac_meth.cpp
	Purpose: Help methods to handle IUPAC letters and comparision

	@author Thomas Büchler
*/

#include <stdexcept>  
#include "iupac_meth.hpp"
using namespace std;


/**
 *  \brief tests whether b is in {A,C,G,T}, or not
 *  \param iupac character b
 *  \return true, if b is in {A,C,G,T}, false otherwise
 */
bool isBase(const char b) {
	return b == 'A' || b == 'C' || b == 'G' || b == 'T';
}

/**
 * \brief maps a string of bases to a IUPAC Letter
 * \param bases containing possible bases in lex order and unice
 * \return the corrosponding IUPAC letter
 */
char getIUPAC_Letter(const string & bases){
  if(bases == "A") 	return 'A';  
  if(bases == "C") 	return 'C';
  if(bases == "G") 	return 'G';
  if(bases == "T") 	return 'T';    
  
  if(bases == "AC") 	return 'M';
  if(bases == "AG") 	return 'R';
  if(bases == "AT") 	return 'W';  
  if(bases == "CG") 	return 'S';
  if(bases == "CT") 	return 'Y';
  if(bases == "GT") 	return 'K';  
  
  if(bases == "ACG") 	return 'V';
  if(bases == "ACT") 	return 'H';
  if(bases == "AGT") 	return 'D';
  if(bases == "CGT") 	return 'B';
  
  if(bases == "ACGT")	return 'N';  
  
  throw std::invalid_argument( "INVALID_BASESTRING in getIUPAC_Letter(): >" + bases + "<" );
}

/**
 * \brief returns the possible iupac letters for a base (base itself is not included)
 */
string getPossibleIUPAC_Letters(const char base){
  switch(base){
    case 'A': return "DHMNRVW";
    case 'C': return "BHMNSVY";
    case 'G': return "BDKNRSV";
    case 'T': return "BDHKNWY";  
    
    case 'B': return "CDGHKMNRSTVWY";  
    case 'D': return "ABGHKMNRSTVWY";
    case 'H': return "ABCDKMNRSTVWY";  
    case 'K': return "BDGHNRSTVWY";  
    case 'M': return "ABCDHNRSVWY";  
    case 'N': return "ABCDGHKMRSTVWY";  
    case 'R': return "ABDGHKMNSVW";  
    case 'S': return "BCDGHKMNRVY";  
    case 'V': return "ABCDGHKMNRSWY";  
    case 'W': return "ABDHKMNRTVY";  
    case 'Y': return "BCDHKMNSTVW";  
  }  
  throw std::invalid_argument( "INVALID_BASE getPossibleIUPAC_Letters(): >" + string(1,base) +"<" );
}

/**
 * \brief returns the possible bases for a iupac letters * 
 */
string getLetters(const char iupac){
  switch(iupac){
    case 'A': return "A";
    case 'B': return "CGT";   
    case 'C': return "C";
    case 'D': return "AGT";   
    case 'G': return "G";
    case 'H': return "ACT";
    case 'K': return "GT";   
    case 'M': return "AC";   
    case 'N': return "ACGT";   
    case 'R': return "AG";   
    case 'S': return "CG";   
    case 'T': return "T";   
    case 'V': return "ACG"; 
    case 'W': return "AT";   
    case 'Y': return "CT";
  }
  throw std::invalid_argument( "INVALID_UPAC_LETTER  getLetters(): >" + string(1,iupac) +"<" );
}

/**
 * \brief returns a grey code number for each iupac letter 
 */
short iupac_number(const char iupac ){  
  /*
   * A 8
   * C 4
   * G 2
   * T 1
   */  
  switch(iupac){
    case 'A': return 8;
    case 'B': return 7;   
    case 'C': return 4;
    case 'D': return 11;   
    case 'G': return 2;
    case 'H': return 13;
    case 'K': return 3;   
    case 'M': return 12;   
    case 'N': return 15;   
    case 'R': return 10;   
    case 'S': return 6;   
    case 'T': return 1;   
    case 'V': return 14; 
    case 'W': return 9;   
    case 'Y': return 5;
 }  
 throw(std::invalid_argument("INVALID_UPAC_LETTER in iupac_number(): >" + string(1,iupac) +"<"));
}

/**
 * \brief returns the iupac letter for a grey code number
 * 
 * inverse of iupac_number()
 */
short iupac_letter(const short num ){  
  /*
   * A 8
   * C 4
   * G 2
   * T 1
   */  
  switch(num){
    case 1: return 'T';
		case 2: return 'G';   
    case 3: return 'K';
    case 4: return 'C';   
    case 5: return 'Y';
    case 6: return 'S';
    case 7: return 'B';   
    case 8: return 'A';   
    case 9: return 'W';   
    case 10: return 'R';   
    case 11: return 'D';   
    case 12: return 'M';   
    case 13: return 'H'; 
    case 14: return 'V';   
    case 15: return 'N';
 }  
 throw(std::invalid_argument("INVALID_UPAC_NUMBER in iupac_letter(): >" + to_string(num) +"<"));
}

/**
 * \brief returns if two chars are matching
 */
bool matches(const char jischar, const char patchar){  
  return( iupac_number(jischar) & iupac_number(patchar));
}

/** \brief returns the complement to a base */
char inverse_char(char & c){
  switch(c){
    case 'A': return 'T';
    case 'T': return 'A';
    case 'G': return 'C';
    case 'C': return 'G';
    case 'N': return 'N';
    
    case 'B': return 'V';   
    case 'D': return 'H';   
    case 'H': return 'D'; 
    case 'K': return 'M'; 
    case 'M': return 'K';     
    case 'R': return 'Y';  
    case 'S': return 'S';    
    case 'V': return 'B'; 
    case 'W': return 'W';    
    case 'Y': return 'R'; 
  }  
  throw invalid_argument("INVALID_UPAC_LETTER in inverse_char(): >" + string(1,c) + "<");
}

 /** \brief selects IUPAC code letter and returns it */
char getSNPLetter(const string & alt, const char ref){
	short number = iupac_number(ref);
	
  for(unsigned int i = 0; i < alt.size(); i++){
    if(alt[i] ==',') continue;
		      number |= iupac_number(alt[i]);
  }
  return iupac_letter(number);
}
