/**
	calc_kmers.cpp
	Purpose: Calculation of the kmer index

	@author Thomas Büchler
*/
#include "../structs/compressedjia_rank.hpp"
#include "../structs/interval_list.hpp"
#include "../structs/kmer_index.hpp"

#include "../methods/io_helper.hpp"

#include "../methods/backwardsearch.hpp"
#include "../methods/construct_methods.hpp"
#include "filenames.hpp"

using namespace sdsl;
using namespace std;

#define TIME double(clock() - time) / CLOCKS_PER_SEC
#define CLEAR_LINE	cout << "\r"<< setw(100) <<" \r" << flush


template <class fmi_type, class jia_type>
kmer_index iteration_step(unsigned int j, kmer_index & last_kmi, fmi_type & fmi, jia_type & jia, string & jis, transform_id & trans){	
	flush_row("calculating k=" + to_string(j),"0%");
	  
	clock_t time = clock();
	
	//for progress bar
	unsigned int prog 		= 0;	
	
	//result vector    
	kmer_index kmi = kmer_index(j);
	
	if(j==1){
		for(unsigned int i = 0; i < kmer_index::letter_count; i++){				
			string pattern = kmer_index::get_string(i,j);
			kmi.kmer[i] = back_search<fmi_type,jia_type>(pattern, fmi, jia, jis, trans, j, true);
		}			
	}else{
		//iterate over calced vectors for k = j-1
		for(unsigned int i = 0; i < last_kmi.kmer.size(); i++){
			string pattern = kmer_index::get_string(i,j-1);
			interval_list ivl = last_kmi.kmer[i];
			//iterate over letters		
			for(unsigned int c = 0; c < kmer_index::letter_count; c++){
				string ext_pattern = kmer_index::int2base(c) + pattern;
				kmi.kmer[c+kmer_index::letter_count*i] = back_search<fmi_type,jia_type>(ext_pattern, fmi, jia, jis, trans, j, true,1,ivl);
			}
			{//Progress status	  
				unsigned int pc = i*100 /  last_kmi.kmer.size();
				if( pc != prog){ 
					prog = pc;
					flush_row("Calculating k=" + to_string(j),to_string(pc)+"%");
				}
			}
		}		
	} 		
	CLEAR_LINE;
	print_row("Time for k=" + to_string(j),TIME);
	flush_row("finish this interation...","");	
	return move(kmi);  
}




/** \brief calculates the interval_list for each read of length k
 * 
 * k = 1: the reads are A,C,G,T: serach each read and store interval_list
 *
 * k > 1 calc precalculation for 1 then for 2 ... till k
 *
 * \param ji_filename filename of ji file
 * \param out_filename filename where precalculation shoud be stored
 * \param k readlength (4^k interval_list will be calculated)
 * 
 */
template <class fmi_type,class jia_type>
void calc_kmer_iterative(int k, string & output,fmi_type & fmi,jia_type & jia, transform_id & trans, string & jis){ 
  
  clock_t time = clock(); 
	
	//iterate j = 1 ... k (dynamic programming)
	//vector<interval_list> last_kmer_vec = vector<interval_list>(0);  //buffer for kmer_vec (k = j)	
	kmer_index last_kmi;
	
	for(int j = 1; j <= k ; j++){
		last_kmi = iteration_step(j,last_kmi,fmi,jia,jis,trans);
	}

	
	
	CLEAR_LINE;
	dotline();
	print_row("Time for kmer_index:",TIME);
  time = clock();
	
	cout << "Store index object to file..." << flush;	    
  if(!store_to_file<kmer_index>(last_kmi, output)){
		CLEAR_LINE;
		print_row("Storing failed", "!!!"); 
    return;
  }
  
	CLEAR_LINE;
	print_row("Time for storing:",TIME);
	dotline();	
	print_row("Kmer index stored in file:",output);
}



/** \brief checks input of main
 * 
 * arg1 = filename of ji
 * arg2 = output
 * 
 * Opt:
 * arg3 = "-k" arg4 = readlength
 * 
 * \param argc 		argc of main
 * \param argv 		argv of main
 * \param wrongcall 	set to true if checking failed
 * \param k 		set if option in args
 * \param output 	name of the output file
 * 
 * following objects are loaded
 * \param fmi 	fm_index
 * \param jia		jump index array
 * \param trans	transformer class
 * \param jis 	jump index string 
 * 
 */
template<class fmi_type,class jia_type>
void checkinput(int & argc,  char** argv, bool & wrongcall, int & k, string & output, fmi_type & fmi,jia_type & jia, transform_id & trans, string & jis){
	clock_t time = clock();

	if(argc < 2 || argc == 3 || argc >= 5){
    cout << "argument count invalid" << endl; 
		wrongcall = true; 
		return;
	}


	output = argv[1] + KMER_FNAME;
	wrongcall |= sdsl_load<fmi_type>(argv[1] + FMI_FNAME,fmi);
	wrongcall |= sdsl_load<jia_type>(argv[1] + JIA_FNAME,jia);
	wrongcall |= sdsl_load<transform_id>(argv[1] + TRANSFORM_FNAME,trans);
	wrongcall |= load_string_from_file(argv[1] + JIS_FNAME,jis);


	for( int i = 2; i < argc; i++ ){
		string arg = argv[i];

		if("-k" == arg)
			k = stoi( argv[++i]);
		else{
			cout << "parameter: '" + arg +"' unknown.\n";
			wrongcall = true;	
		}
	}
	print_row("Time for loading fmi, jia, jis, trans:",TIME);
}


/**
 * \brief precalculates interval_list for all reads of length k 
 */
int main(int argc,  char** argv) {  
	typedef sdsl::csa_wt<sdsl::wt_huff<sdsl::bit_vector,sdsl::bit_vector::rank_1_type,sdsl::select_support_scan<1>,sdsl::select_support_scan<0>>,32,1024*1024>	fmi_type;
	typedef compressedjia_rank jia_type;
	clock_t time = clock();
	
  bool wrongcall	= false;
  int k 		= 10;
  string output;
	fmi_type fmi;
	jia_type jia;
	transform_id trans;
	string  jis;
  
  checkinput<fmi_type,jia_type>(argc, argv, wrongcall,k, output,fmi,jia,trans,jis);     
  
  if(wrongcall){	
    cout << "Usage of " << argv[0] << endl << "1 parameter needed" << endl;
    cout << " arg1 = output dir of construct_index (including " << FMI_FNAME << " " << JIA_FNAME << " " << JIS_FNAME << " " << TRANSFORM_FNAME << ")" << endl;
    cout << "Optional param:\n -k length of the kmers (4^k reads will be calculatet, default is k=10)" << endl;
		cout << "This program calculates a kmer index" << endl;
    return 0;    
  }  
     
	dotline();
	print_row("Calculate kmer index of ",argv[1]);
  cout << "upto k=" << k << "." << endl;
	dotline();
  calc_kmer_iterative<fmi_type,jia_type>(k, output, fmi,jia,trans,jis);
	print_row("Time in total:",TIME);
  return 0;
}
