/**
	construct_index2.cpp
	Purpose: Construction of the jump index

	@author Thomas Büchler
	@version 2 (with copy number variaion)
*/

#include <sys/stat.h> //mkdir
#include <iostream>	//cout
#include <unistd.h> //file test


#include "../methods/construct_jis.hpp"
#include "../methods/io_helper.hpp"

#include "../structs/sequence_streams.hpp"
#include "../structs/construct_streams.hpp"
#include "../structs/compressedjia_select.hpp"
#include "../structs/compressedjia_rank.hpp"
#include "../structs/jumpindexarray.hpp"
#include "../structs/transform.hpp"



#include <sdsl/int_vector.hpp>
#include "filenames.hpp"

#define TIME double(clock() - time) / CLOCKS_PER_SEC


using namespace std;
using namespace sdsl;


/**
 * \brief builds the jump index
 * 
 * steps:
 * 
 * -build index string and index array
 * -build fm index
 * -calculate transformation (numbering by rank in jis to numbering by rank in bwt)
 * -transform numbering of index array
 * -compress index array
 * -build jump index
 * -store jump index and transform
 * 
 * \param fa_filename
 * \param vcf_filename
 * \param out_dirname 
 */
template<class fmi_type, class jia_type>
void build_index(string fa_filename, string vcf_dirname, string out_dirname ){ 
    
  //declare timer
  clock_t	time0 	= clock();  
  

  
  //varaibles for the index construction
  sequence_streams	seq_s;		// streams to sequence data
  construct_streams con_s;		// streams to other data  
  fmi_type 	fmi;						// fm_index containing: bwt-rank_support, C, (SA (sampled),ISA (no samples))
  transform_id trans;						// struct to transform numbering of the markers: bwt_rank <-> jis_rank <-> jis_position	
  jumpindexarray jia;					// jumpindex array  
	jia_type 	cia;	// compressed jia 
	bool 		storing_failed	= false; //bool that indicates problems with storing to file
  

    
  
  //open filestreams
	seq_s	= sequence_streams(fa_filename, vcf_dirname, out_dirname);
	con_s = construct_streams(out_dirname);

	
	//construct index string
	
	
	
	//construct index string
	clock_t time = clock();
	jia =  construct_jis(seq_s, con_s);
	print_row("Time for index string", TIME); time = clock();
	
	//build fm index
	flush_row("buliding FM index","...");
  construct(fmi, seq_s.jis_filename, 1);
	print_row("Time for FM index", TIME); time = clock();
  
   
	//calc transformations	
	flush_row("calculating transform permutation","...");
	trans = transform_id::factory<fmi_type>(fmi); 		
	print_row("Time for transform permutation", TIME); time = clock();
	
	
	//store fm index
	flush_row("storing FM index","...");
	string fmi_filename = out_dirname + FMI_FNAME;
	if(!sdsl::store_to_file(move(fmi),fmi_filename)){
		storing_failed 	= true;
		con_s.log 		<< fmi_filename << " WAS NOT STORED CORRECTLY" << endl;   
	}
	print_row("Time for storing FM-index", TIME); time = clock();
	
	
	//tramsform jia
	flush_row("transforming jia","...");
	jia = jia.get_transformed_jia(trans.rs2rb);
	print_row("Time for transforming jia",TIME); time = clock();
		

	//compress jia
	flush_row("compressing jia","...");
	cia 	= jia_type(jia); 	
	print_row("Time for compressing jia",TIME); time = clock();
	
	
	//store compressed jia
	flush_row("storing compressed jia","...");
	string jia_filename = out_dirname + JIA_FNAME;
	if(!sdsl::store_to_file(move(cia),jia_filename)){
		storing_failed 	= true;
		con_s.log 		<< jia_filename << " WAS NOT STORED CORRECTLY" << endl;   
	}
	print_row("Time for storing compressed jia",TIME); time = clock();
	
	
	//store trans
	
	flush_row("storing transformation","...");
	string trans_filename = out_dirname + TRANSFORM_FNAME;
	if(!sdsl::store_to_file(move(trans),trans_filename)){
		storing_failed 	= true;
		con_s.log 		<< trans_filename << " WAS NOT STORED CORRECTLY" << endl;   
	}	  
	print_row("Time for storing transformation",TIME); time = clock();
	
	// write infos in cout   
	
	dotline();
	print_row("Time in total:", double(clock() - time0) / CLOCKS_PER_SEC);	
	print_row("Please notice logfile:" , con_s.log_filename); 	
	if(con_s.error)		
		print_row("Please notice logfile:" , con_s.log_filename+"error"); 	
	dotline();
	
	if(storing_failed)
		print_row("Storing failed", "!!!");   
  
  con_s.close();
}



/** \brief checks input of main
 * 
 * arg1 = fa_filename
 * arg2 = vcf_dirname 
 * arg3 = out_dirname
 * 
 * test if following exist:
 * -fa_filename
 * -vcf_dirname/snp
 * -vcf_dirname/indel
 * -vcf_dirname/sv
 * 
 * test if out_dirname exist or can be created
 * 
 * 
 * \param argc 		argc of main
 * \param argv 		argv of main
 * \param wrongcall	set to true if checking failed
 * 
 */
void checkinput(int & argc,  char** argv, bool & wrongcall){

  if(argc != 4){
    cout << "argument count invalid" << endl; 
    wrongcall = true; 
    return;
  }    
	
	wrongcall |= file_access(argv[2] + SNP_FNAME);
	wrongcall |= file_access(argv[2] + IND_FNAME);
	wrongcall |= file_access(argv[2] + SV_FNAME);
	wrongcall |= file_access(argv[1]);

  	 
  if(mkdir(argv[3], 0777) != 0){	//error if dir could not be created
    if(errno != EEXIST){
			cout << argv[3] << " could not be created!" << endl;
			wrongcall = true; 			
		}
  } 

}



/**
 * \brief checks arguments and calls parseVCF()
 * 
 */
int main(int argc,  char** argv) {
	typedef sdsl::csa_wt<sdsl::wt_huff<sdsl::bit_vector,sdsl::bit_vector::rank_1_type,sdsl::select_support_scan<1>,sdsl::select_support_scan<0>>,32,1024*1024>		fmi_type;
	typedef compressedjia_rank jia_type;
  
  //input
  bool 		wrongcall 		= false;
  
	checkinput(argc, argv, wrongcall);
	
  if(wrongcall){  //usage ouput
    cout << "Usage of " << argv[0] << endl <<"3 parameter needed:" 		<< endl;
    cout << " fa-filename\n vcf preparation directory (where result of prepare_vcf are stored)\n output directory" 	<< endl;
    
    cout 									<< endl;
    cout << "This program constructs the 'jump index' to the sequence" << endl;
    return 0;    
  }  
  	//write info
	dotline();
	print_row("Construct index of", argv[1]);
	print_row("with variation of", argv[2]);
	dotline();
	
	
  build_index<fmi_type, jia_type>(argv[1],  argv[2],  argv[3]);
}

