/**
	match.cpp
	Purpose: Read input and call backward search

	@author Thomas Bchler
*/
#include <iostream> //cout
#include <iomanip> //setw
#include <fstream> // ifstream (fastq)
#include <unistd.h> //file test
#include <string>
#include <list>

#include "../structs/searchinterval.hpp"
#include "../structs/interval_list.hpp"
#include "../structs/kmer_index.hpp"
#include "../structs/searchresult.hpp"
#include "../structs/matchresult.hpp"

#include "../structs/jumpindexarray.hpp"
#include "../structs/compressedjia_select.hpp"
#include "../structs/compressedjia_rank.hpp"

#include "../methods/backwardsearch.hpp"
#include "../methods/io_helper.hpp"
#include "filenames.hpp"

using namespace sdsl;
using namespace std;


/** \brief calculates the 5'-3' reverse of a DNA pattern*/
std::string inverse_pattern(std::string & pattern){
  string inv = string(pattern);
  for(unsigned int i = 0; i < pattern.length() ; i++)
    inv[i] = inverse_char(pattern[pattern.length() - i - 1]);
  return(move(inv));
}


/** \brief matches the pattern in the fastq file
 * 
 * fastq parsing
 * calc of inverse pattern
 * 
 * patternmatching
 */
template <class fmi_type, class jia_type>
void startmatcher(const string & fastq_filename, const string & out_filename, bool use_kmer, int maxN, 
									fmi_type & fmi, jia_type jia, string & jis, transform_id & trans, kmer_index & kmi){  
  
	//open log
  string log_filename		= out_filename + "_log";
  ofstream log;
  log.open(log_filename);
  
	//timer measure
  clock_t time 		= clock();
  
    
  //storage for output
  matchresult result;
  
  
  //open input stream
  ifstream pattern_stream;
  pattern_stream.open(fastq_filename);
  
  //counters
  unsigned int skipped = 0;	//number of skipped reads
  unsigned int count		= 0;	//number of reads
  
  //loop over fastq entries
  while(true){    
    string 	name; 
    string 	pattern; 
    string 	qual;          
    
    { //parse a fastq entry
      if( ! getline(pattern_stream, name) ) 	break;  //name: @.... 
      if( ! getline(pattern_stream, pattern) ) 	break;  //pattern: AAAAAAATT...
      if( ! getline(pattern_stream, qual) ) 	break;  //+
      if(qual != "+"){
				log << name << ": Invalid FASTQ. No '+' line."  << endl; 
				skipped++;
				getline(pattern_stream, qual);
				continue;
      }
      if( ! getline(pattern_stream, qual) ) 	break;  //qual: H25'*!!HH..
      if(qual.length() != pattern.length()){
				log << name << ": Invalid FASTQ. Pattern and Qual have to be of same length." << endl;
				skipped++;
				continue;
      }
    }
    
    
    // i = 0 -> search pattern
    // i = 1 -> search inverse
    for(int i = 0; i < 2; i++){
      if(i == 1){ // calc inverse of pattern
				pattern	= inverse_pattern(pattern);	
				name 	+= "_inv";
				//TODO if qual should be used, qual has to be reversed. Qual is not used in this project yet
      }
      interval_list 	intervals;
      
      if(use_kmer){//calc interval_list with kmer index           
				int k = kmi.k;
				
				if(k 	>= (int )pattern.length()){
					log << name << ": Pattern to short for used kmer index" << endl;
					skipped++;
					continue;
				}
		
				//calc tail - kmer of strings
				const string kmer 		= pattern.substr(pattern.length()-k);
		
				try{
					//load results of this kmer
					interval_list kmer_list	= kmi.get_ivlist(kmer);
					//search
					intervals			= back_search<fmi_type,jia_type>(pattern, fmi, jia,  jis, trans, maxN, false, pattern.length()-k, kmer_list);  
				}catch (invalid_argument & ia){
					log << ia.what()<< endl;
					log << pattern << endl;
					log << name << ": Pattern contains a unallowed letter for kmer index (prob. 'N')" << endl;
					skipped++;
					continue;
				}
				
      }else{//calc interval_list without kmer index      
				intervals 			= back_search<fmi_type,jia_type>(pattern, fmi, jia,  jis, trans, maxN);       
      }          
      
      searchresult res 			= searchresult(name,pattern,qual,move(intervals));      
      result.add_res(move(res));
      
      count 	+= 1;
      if(count % 10000 == 0){
				flush_row("Reads:", to_string(count));
			}
    }
  }
  cout << '\r';   
  pattern_stream.close();
  log.close();
  print_row("Skipped reads", skipped); 
  print_row("Searched reads", (unsigned int) result.match_res.size()); 
  print_row("Mapped reads", result.mapped()); 
  print_row("Results", result.results());
	dotline();	
	print_row("Time overall: ",((double)(clock()-time) / CLOCKS_PER_SEC));   
  if(skipped > 0){
		print_row("Notice logfile", log_filename); 
  }

  if(!store_to_file<matchresult>(result, out_filename) )
		print_row("Storing failed", "!!!"); 
  else
		print_row("Results  stored in file:", out_filename);
  
}

/** \brief checks and loads input of main
 * 
 * arg1 = filename of buid dir
 * arg2 = filename of fastq
 * arg3 = output 
 * 
 * Opt:
 * -prec precalculation-file
 * 
 * \param argc 		argc of main
 * \param argv 		argv of main
 * \param wrongcall	set to true if checking failed
 * \param use_kmer	set if build contains file 'kmer' 
 * 
 * following objects are loaded
 * \param fmi 	fm_index
 * \param jia		jump index array
 * \param trans	transformer class
 * \param jis 	jump index string 
 * \param kmi 	kmer index
 */
template<class fmi_type, class jia_type>
void checkinput(int & argc,  char** argv, bool & wrongcall, bool & use_kmer, fmi_type & fmi, jia_type & jia, string & jis, transform_id & trans, kmer_index & kmi){
	clock_t time = clock();
	
	if(argc != 4){
    cout << "argument count invalid" << endl; 
		wrongcall = true; 
		return;
	}  
	//load files
	wrongcall |= sdsl_load<fmi_type>		(argv[1] + FMI_FNAME, fmi);
	wrongcall |= sdsl_load<jia_type>		(argv[1] + JIA_FNAME, jia);
	wrongcall |= sdsl_load<transform_id>(argv[1] + TRANSFORM_FNAME, trans);
	wrongcall |= load_string_from_file	(argv[1] + JIS_FNAME, jis);
	wrongcall |= file_access(argv[2]);


  string kmer_filename	= argv[1] + KMER_FNAME;
	use_kmer = !access(kmer_filename.c_str(), F_OK);
	if(use_kmer)
		wrongcall |= sdsl_load<kmer_index>(kmer_filename,kmi);
	
	print_row("Time for loading fmi, jia, jis, trans, (kmer):", (double)time/ CLOCKS_PER_SEC);
	dotline();
}

/**
 * \brief match a pattern to the jumpindex
 * 
 * calls checkinput and startmatcher
 */
int main(int argc,  char** argv) {  
	typedef sdsl::csa_wt<sdsl::wt_huff<sdsl::bit_vector,sdsl::bit_vector::rank_1_type,sdsl::select_support_scan<1>,sdsl::select_support_scan<0>>,32,1024*1024>	fmi_type;
	typedef compressedjia_rank jia_type;
	
  bool wrongcall 	= false;
  bool use_kmer 	= false;
  //number of allowed N in matching
	int maxN = 5; //TODO may as input
	
	fmi_type fmi;
  jia_type jia;  
  transform_id trans;
  string jis; 
  kmer_index kmi;
  
	checkinput<fmi_type,jia_type>(argc, argv, wrongcall,use_kmer, fmi, jia, jis, trans, kmi);   		

	
  if(wrongcall){
    cout << "Usage of " << argv[0] << endl << "3 parameter needed" << endl;
    cout << " output dir of construct_index (including " << FMI_FNAME << " " << JIA_FNAME << " " << JIS_FNAME << " " << TRANSFORM_FNAME <<" [" << KMER_FNAME << "])" << endl;		
		cout << " fastq file\n result file (output)" << endl;
    cout << "If build directory contains a file '"<<KMER_FNAME<<"', this will be used as kmer index" << endl;
    return 0;    
  }  
  
// 	//write info
	print_row("Match", argv[2]);
	print_row("in", argv[1]);
	cout << "with" << (use_kmer?"":"out") << " using kmer index";
	use_kmer?(cout << " k=" << kmi.k << endl):cout << endl;
	dotline();

	
	//remove_n entries in kmi where 'N' matched more often then maxN times. 
	// this "filter" may is not needed, because in the calculation of the kmi this filter is also used
	// when maxN gets a input this gets relevant again
	kmi.remove_n(maxN);
	
  startmatcher<fmi_type,jia_type>(argv[2],argv[3],use_kmer, maxN, fmi, jia, jis, trans, kmi);
}
