/**
	prepare_VCF.cpp
	Purpose: Read input and call backward search

	@author Thomas Büchler
*/

#include <iostream>
#include <string>
#include <fstream>
#include <ctime>
#include <unistd.h>
#include <stdexcept>
#include <sstream> 
#include <sys/stat.h> //mkdir
#include <deque> 
#include "../methods/io_helper.hpp"
#include "../structs/vcfl.hpp"
#include "../structs/conflict_group.hpp"
#include "../structs/sequence_streams.hpp"


using namespace std;

#define CONFLICT_TXT "\tCONFLICT\tgenerating combinations"
#define REPEAT_TXT "\tIGNORED\trepetitive enty"
#define UNKNOWN_TXT "\tIGNORED\tonly SNPs INDELS and CNV treated"
#define OVERLAP_TXT1 "\tIGNORED\t"
#define OVERLAP_TXT2 "overlapping with more frequent SV"


/**
 * \brief searches for 'AC=x,AC=y,...' in info string and returns max(x,y...)
 */
unsigned int sv_maxAC(const vcfl & sv){
	size_t ac_pos1 = sv.info.find("AC=")+3;
	int max_ac = 1;
	
	for(int i = 1; i < sv.altCount(); i++){
		size_t ac_pos2 = sv.info.find(',',ac_pos1)+1;
		int ac = stoi(sv.info.substr(ac_pos1,ac_pos2-ac_pos1));
		if(ac > max_ac) max_ac=ac;
		ac_pos1 = ac_pos2;		
	}
	//same with ; instead of , because last one ist seperated differently
	size_t ac_pos2 = sv.info.find(';',ac_pos1)+1;
	int ac = stoi(sv.info.substr(ac_pos1,ac_pos2-ac_pos1));
	if(ac > max_ac) max_ac=ac;
	
	return max_ac;
}

/**
 * \brief reads VCF file and sorts entries
 * 
 * SNPs   will be written to a seperate file
 * INDELs   -"-
 * CNVs     -"-
 * 
 * Collisions of INDELs and SNPs will be solved trough generating all cominations and adding IUPAC encoding of SNP to ref/alt of INDEL.
 * 
 * Other SVs as CNV are not supportet at the moment
 * 
 * When CNV overlap the one with the highest AC is choosed 
 * 
 */
void prepareVCF(string & vcf_filename, string & out_dir){
	
	string dummy = "";
	sequence_streams seq_s = sequence_streams(dummy, out_dir, dummy);
	

  clock_t t = clock();
  
  
  //open stream
  ifstream in_vcf;	
  ofstream out_stat;
  ofstream log;
  
  in_vcf		.open(vcf_filename);
	seq_s.open_for_prepvcf(vcf_filename);
	out_stat	.open(out_dir+"/prepare_stats");
	log	      .open(out_dir+"/prepare_log");	

	log 			<< "## Log of preparation of file: " << vcf_filename<< endl;
	log 			<< "#Line\tEVENT   \texplanation" << endl;
	
	
  //read vcf file
  string line;
  unsigned int linecount = 0; 
  

  int snp_count 	= 0;
  int indel_cout 	= 0;
  int sv_count 		= 0;
  int sitemarker_count 	= 0;
  int seperator_count 	= 0;
  int rep_count 	= 0;
  int ign_sv_count 	= 0;
	
	vcfl last_v = vcfl();
	
	vcfl sv_candidate = vcfl(); 
	// if sv_candidate is not in conflict with other SV (overlapping), then its written to out_sv
	// if its in conflict with the SV sv then the entry with higher AC (allele count) will be the next candidate
	
	conflict_group cogru = conflict_group();
    
  while(getline(in_vcf, line)){  
		//ignore comments
    linecount++;    
    if(linecount % 100000 == 0)
			flush_row("Lines:",to_string(linecount));
    if(line[0] == '#')		continue;
    
    vcfl v = vcfl(line);		
		if(v.equal(last_v)){	//same variation as in line before
			log << linecount << REPEAT_TXT <<endl;
			rep_count ++;
			continue;
		}
		
		if(v.isINDEL()){ //is true even if SNP. SNP and INDELS are added to cogru
			
			if(cogru.conflict(v)){ //when conflict then added to cogru
				cogru.add(v);
				log << linecount << CONFLICT_TXT <<endl;
				
			}else{ 
				//default case: write variation of cogru to file
				if(!cogru.empty()){
					vcfl vc = cogru.solve();
					if(vc.isSNP()){ // write SNP in file
						seq_s.snp_out << vc;
						snp_count++;
					}else{// write INDEL in file
						seq_s.ind_out << vc;
						indel_cout++;
						sitemarker_count++;
						seperator_count += vc.altCount() +1;
					}
				}
				cogru.clear();
				cogru.add(v);				
			}
			
		}else{//SV treatment
			if(v.alt.substr(0,3) != "<CN"){				
				log << linecount << UNKNOWN_TXT <<endl;
				ign_sv_count++;
			}else{//is CNV
				if(sv_candidate.empty()){ //set v as candidate
					sv_candidate = v;
				}else{
					
					if(sv_candidate.sv_end() < v.pos || sv_candidate.chrom != v.chrom){ // no conflict sv_candidate can be used, v is next candidate
						seq_s.sv_out << sv_candidate;
						sv_count++;
						seperator_count++;
						sitemarker_count += sv_candidate.maxCN();
						sv_candidate = v;
					}else{
						ign_sv_count++;
						if(sv_maxAC(sv_candidate) >= sv_maxAC(v)){//candidate wins, because of higher AC
							log << linecount << OVERLAP_TXT1  << v.id << OVERLAP_TXT2 << sv_candidate.id << endl;
						}else{// v wins
							log << linecount << OVERLAP_TXT1  << sv_candidate.id << OVERLAP_TXT2 << v.id << endl;
							sv_candidate = v;
						}					
					}				
				}
			}
		}		
		last_v = v;    
  }
  
  //write entries in buffers to file
  
  if(!cogru.empty()){
		vcfl vc = cogru.solve();
		if(vc.isSNP()){ // write SNP in file
			seq_s.snp_out << vc;
			snp_count++;
		}else{
			seq_s.ind_out << vc;
			indel_cout++;
			sitemarker_count++;
			seperator_count += vc.altCount() +1;
		}
	}
	
	if(!sv_candidate.empty()){
		seq_s.sv_out << sv_candidate;
		sv_count++;
		seperator_count++;
		sitemarker_count += sv_candidate.maxCN();
	}
	
  seq_s.close();
  in_vcf.close();
	log.close();
	seperator_count++; //between ref and cnv sections
	out_stat << snp_count 		<< "\tSNPs"				<< endl;
	out_stat << indel_cout		<< "\tINDELs"			<< endl;
	out_stat << sv_count			<< "\tused SVs"		<< endl;
	out_stat << ign_sv_count	<< "\tignored SVs"<< endl;
	out_stat << sitemarker_count	<< "\tsite markers"	<< endl;
	out_stat << seperator_count		<< "\tseperators"		<< endl;
	out_stat << rep_count			<< "\trepetitive lines"	<< endl;
	
	out_stat.close();
	print_row("Lines: ",linecount);
	print_row("Time needed: ",(double(clock() - t))/ CLOCKS_PER_SEC);
  
}

/** \brief checks input of main
 * 
 * arg1 = filename of vcf
 * arg2 = output 
 * 
 * test if files: vcf exist
 * 
 * 
 * \param argc 		argc of main
 * \param argv 		argv of main
 * \param wrongcall	set to true if checking failed
 * 
 */
void checkinput(int & argc,  char** argv, bool & wrongcall){

  if(argc != 3){
    cout << "argc " << argc << endl; 
    wrongcall = true; 
    return;
  }    
  string vcf_filename 	= argv[1];
  if(access((vcf_filename).c_str(), F_OK)){
    cout << "file '" << vcf_filename << "' not found!" << endl << endl;
    wrongcall = true; 
  }   
  
	if(mkdir(argv[2], 0777) != 0){	//error if dir could not be created
    if(errno != EEXIST){
      cout << argv[2] << " could not be created.ls. [ERROR CODE=" <<  to_string(errno) <<" see errno.h]";
			wrongcall = true;
		}
  }
}


/**
 * \brief checks inputs and starts programs
 */
int main(int argc,  char** argv) {  

	bool wroncall;
	checkinput(argc, argv, wroncall);	
  if(wroncall){    
    cout << "Usage of " << argv[0] << endl <<"2 parameter needed:" << endl;
    cout << " vcf-file\n output directory"  << endl;
		cout << "This programs splits the entries of the vcf file in SNPs, INDELs and SVs and write them in separat files" <<endl;
    return 0;
  }

  string vcf = string(argv[1]);
	string out = string(argv[2]);
	dotline();
	print_row("Preparing VCF file", argv[1]);
	dotline();
  prepareVCF(vcf,out);    
    
}
  



