/**
	print_graph.cpp
	Purpose: print graph of pan genome in dot (gfa) format

	@author Kadir Dede, Thomas Büchler
*/
#include <iostream>
#include <fstream>
#include <string>
#include <iomanip> 
#include <sstream>      // std::stringstream, std::stringbuf

#include <sdsl/suffix_arrays.hpp>
#include <sdsl/wt_huff.hpp>
#include <sdsl/io.hpp>

#include "../structs/jumpindexarray.hpp"
#include "../structs/compressedjia_select.hpp"
#include "../structs/compressedjia_rank.hpp"
#include "../structs/transform.hpp"
#include "../methods/io_helper.hpp"
#include "../structs/position_helper.hpp"

#include "filenames.hpp"



using namespace sdsl;
using namespace std;

#define FORMAT_DOT 0
#define FORMAT_GFA 1

struct node {
	unsigned int number;
	unsigned int ref_succ;
	unsigned int start;
	unsigned int offset;
};


string cut_middle(string& s, unsigned int max_chars) {
	if(max_chars > 0) {
		if(s.length() <= max_chars) {
			return s;
		} else {
			return s.substr(0, max_chars / 2) + "..." + s.substr(s.length() - max_chars / 2, max_chars / 2);
		}
	} else {
		return s;
	}
}

void graph_print_head(short format, ofstream & graph_out){
	switch (format){
		case FORMAT_DOT:
			graph_out << "digraph G {" << endl;
			graph_out << "  rankdir = \"LR\"" << endl;
			graph_out << "  node [shape = box]" << endl;
      break;
		case FORMAT_GFA:
			graph_out << "H\tVN:Z:1.0"<< endl;
			break;
		}	
}

void graph_print_node(short format, ofstream & graph_out, string node, string label){
		switch (format){
			case FORMAT_DOT: graph_out << "  " << node << " [label = \"" << label << "\"]" << endl;	break;
			case FORMAT_GFA: graph_out << "S\t" << node << "\t" << label << endl;
		}	
}

void graph_print_edge(short format, ofstream & graph_out, string a, string b){
		switch (format){
			case FORMAT_DOT: graph_out << "  " << a << " -> " << b << endl;	break;
			case FORMAT_GFA: graph_out << "L\t" << a << "\t+\t" << b << "\t+\t0M\n";	break;
		}
}

void graph_print_tail(short format, ofstream & graph_out){
		switch (format){
			case FORMAT_DOT: graph_out << "}" << endl;	break;
			case FORMAT_GFA: break;
		}	 
}


/** \brief prints the dot representation of the jis + jia
 * 
 */
template <class jia_type>
void startreader(const string& out_filename, jia_type& jia, transform_id& trans, string& jis, size_t start, size_t end, unsigned int max_chars, short format) {  

	clock_t time = clock();
	clock_t time0 = time;   

	//open stream
	ofstream graph_out;
	graph_out.open(out_filename);
	graph_print_head(format,graph_out);


	vector<unsigned int> visited = vector<unsigned int>(); // visited nodes
	vector<unsigned int> ref_nodes = vector<unsigned int>(); // nodes to be tunneled
	unsigned int pos = end; // start from back
	bool hit_start = false; // indicates if start node was reached from end

	jumpindexarray jia_uncomp = jia.jia();
	jia_uncomp = jia_uncomp.get_transformed_jia(trans.rb2rs);
	compressedjia_rank jia_c = compressedjia_rank(jia_uncomp);

	stack<node> s = stack<node>(); // stack of nodes to be processed yet
	
	while(pos != start && jis[pos] != '#') {
		pos--;
	}

	node c; // current node
	c.number = trans.rb2rs[trans.ps2rb[pos]] + 1; // number is k-th # in jis
	c.start = (pos == start) ? pos : pos + 1;
	c.offset = end - pos;

	s.push(c);

	while(!s.empty()) {
		c = s.top();
		s.pop();
		if(find(visited.begin(), visited.end(), c.number) != visited.end()) {
			// nothing to do if visited already
			continue;
		}
		visited.emplace_back(c.number);

		bool is_ref_node = false; // current node is a reference node
		if(find(ref_nodes.begin(), ref_nodes.end(), c.number) != ref_nodes.end()) {
			is_ref_node = true;
		}
		
		bool is_start_node = false; // current node is the start node
		if(c.start <= start && start <= c.start + c.offset) {
			is_start_node = true;
			hit_start = true;
		}

		if(!is_ref_node) { // only print node if it is no reference node
			
			string label;			
			if(is_start_node) // print only from start if it is a start node
				label = jis.substr(start, c.offset - (start - c.start));
			else
				label = jis.substr(c.start, c.offset);			
			
			graph_print_node(format, graph_out, to_string(c.number), cut_middle(label, max_chars));
		}

		if(!is_start_node) { // start node has no predecessors at all
			// get targets from previous #
			int_vector<0> targets = jia_c.targets(c.number - 1);
			if(targets.empty() && !hit_start) {
				// no predecessors found and start node not processed yet
				node a;
				a.number = c.number - 1;
				a.start = trans.rs2ps[c.number - 2] + 1;
				a.offset = trans.rs2ps[c.number - 1] - a.start;
				s.push(a);
			}

			for(unsigned int t : targets) {
				bool is_ref_pred = false; // predecessor is a reference node

				if(find(visited.begin(), visited.end(), t) == visited.end()) {
					node a;
					a.number = t;
					a.start = (t == 0) ? 0 : trans.rs2ps[t - 1] + 1;
					a.offset = trans.rs2ps[t] - a.start;
					if(a.offset == 0) { // predecessor is a reference node
						ref_nodes.emplace_back(a.number);
						a.ref_succ = c.number;
						is_ref_pred = true;
					}
					s.push(a);
				}

				if(!is_ref_node && !is_ref_pred)  // if none is reference node
					graph_print_edge(format,graph_out,to_string(t),to_string(c.number));
				else if(is_ref_node && t >= c.ref_succ) // only plot real SVNs!
					graph_print_edge(format,graph_out,to_string(t),to_string(c.ref_succ));
			}
		}
	}

	graph_print_tail(format, graph_out);
	
	print_row("Time overall: ", ((double)(clock() - time0) / CLOCKS_PER_SEC));
	print_row("Output: ", out_filename);
}

/** \brief checks input of main
 * 
 * arg1 = name of build directory
 * arg2 = name of output file 
 * 
 * Opt:
 * 
 * \param argc 		argc of main
 * \param argv 		argv of main
 * \param wrongcall	set to true if checking failed
 * 
 * following objects are loaded
 * \param jia		jump index array
 * \param trans	transformer class
 * \param jis 	jump index string 
 * 
 */
template<class jia_type>
void checkinput(int& argc,  char** argv, bool& wrongcall, jia_type& jia, transform_id& trans, position_helper& pos_h, string& jis, size_t& start, size_t& end, unsigned int& max_chars, short& format) {

	clock_t time = clock();

	//1 = build dir
	//2 = out_file
	if(argc < 3) {
		cout << "argument count invalid" << endl; 
		wrongcall = true; 
		return;	
	}
	
	//load structs
	wrongcall |= sdsl_load<jia_type>(argv[1] + JIA_FNAME, jia);
	wrongcall |= sdsl_load<transform_id>(argv[1] + TRANSFORM_FNAME, trans);
	wrongcall |= sdsl_load<position_helper>(argv[1] + POSITION_FNAME, pos_h);
	wrongcall |= load_string_from_file(argv[1] + JIS_FNAME, jis);
	
	if(wrongcall)
		return;
	
	start = 0;
	end = trans.rs2ps[pos_h.seq_seperator[pos_h.seq_seperator.size()-2]]-1; //last letter of jis before cnv section
	
	try{
		for( int i = 3; i < argc; i++ ){
			string arg = argv[i];
			
			if("-js" == arg){
				size_t new_start = stoi( argv[++i]);
				if(new_start <= end)
					start = new_start;
				else
					cout << "paramater -js ignored (value to big)\n";
			}
			else if("-je" == arg){
				size_t new_end = stoi( argv[++i]);
				if(new_end <= end)
					end = new_end;
				else
					cout << "paramater -je ignored (value to big)\n";
			}
			else if("-max" == arg){
				max_chars = stoi( argv[++i]);			
			}
			else if("-gfa" == arg){
				format = FORMAT_GFA;		
			}
			else{
				cout << "parameter: '" + arg +"' unknown.\n";
				wrongcall = true;	
				return;
			}
		}
	}catch(invalid_argument& ia) {
		wrongcall = true;
		return;
	}	


	print_row("Time for loading jia, jis, transform:",((double)(clock() - time) / CLOCKS_PER_SEC));
	dotline();
}


/**
 * \brief starts forward reader to print GFA graph
 * 
 * calls checkinput and startreader
 */
int main(int argc,  char** argv) {  
	typedef compressedjia_rank	jia_type;

	bool wrongcall 	= false;
	
	jia_type jia;
	transform_id trans;
	position_helper pos_h;	
	string jis;
	
	size_t start;
	size_t end;
	unsigned int max_chars = 0; // max chars to print in a node
	short format = FORMAT_DOT;

	checkinput(argc, argv, wrongcall, jia, trans, pos_h, jis, start, end, max_chars, format); 

	if(wrongcall) {
		cout << "Usage of " << argv[0] << endl << "2 parameters needed" << endl;
		cout << " build-directory (including " << JIA_FNAME << " " << JIS_FNAME << " " << TRANSFORM_FNAME << " " << POSITION_FNAME <<")" << endl;
		cout << " output file" << endl;
		cout << " Optional parameters:" << endl;
		cout << " -js start position of subgraph (in jis)" << endl;
		cout << " -je end position of subgraph (in jis)" << endl;
		cout << " -max maximal count of bases per label " << endl;
		cout << " -gfa a .gfa representation is printed instead of .dot" << endl;
		cout << "This program prints a dot representation of the implicit graph." << endl;
		return 0;    
	}
  
	startreader<jia_type>(argv[2], jia, trans, jis, start, end, max_chars, format);
}

