/**
	write_positions.cpp
	Purpose: Read matchresult and call forward search to genereate readable information

	@author Thomas Büchler
*/


#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <iomanip> 
#include <sstream>      // std::stringstream, std::stringbuf

#include "../structs/searchinterval.hpp"
#include "../structs/interval_list.hpp"
#include "../structs/searchresult.hpp"
#include "../structs/matchresult.hpp"

#include "../structs/jumpindexarray.hpp"
#include "../structs/compressedjia_rank.hpp"
#include "../structs/transform.hpp"
#include "../structs/position_helper.hpp"

#include "../methods/io_helper.hpp"

#include "../methods/construct_methods.hpp"
#include "../methods/forwardsearch.hpp"

#include "filenames.hpp"



using namespace sdsl;
using namespace std;




/** \brief matches the pattern in the fastq file
 * 
 * io handling
 * 
 * fastq parsing
 * calc of inverse pattern
 * 
 * patternmatching
 * 
 * \param out_filename file in wich output is written
 * \param only_mapped if true only mapped reads are written
 */
template<class fmi_type, class jia_type>
void startreader(const string &  result_filename, const string &  out_filename, const bool only_mapped, fmi_type & fmi, jia_type & jia, matchresult & res, transform_id & trans, position_helper & pos_h, string & jis){  


	clock_t time = clock();
	
  //invert jia to 
	//uncompress
	jumpindexarray jia_uncomp = jia.jia();
	//invert
	jia_uncomp = jia_uncomp.get_transformed_jia(trans.rb2rs);
	jia_uncomp = jia_uncomp.get_inverted_jia();
	//compress
	compressedjia_rank jia_c = compressedjia_rank(jia_uncomp);
	print_row("Time for transforming und inverting jia: ",((double)(clock()-time) / CLOCKS_PER_SEC));
  time = clock();
	
  
  unsigned int mapped  = res.mapped();
  unsigned int results = res.results();
  unsigned int twice_mapped = res.twice_mapped();
  
	
	 //open stream
  ofstream out;
  out.open(out_filename);      
  //write info
  out << "# Result file:" << result_filename << endl;
	if(only_mapped)
  out << "# Number of reads: " << res.match_res.size() << endl;
  out << "# Number of mapped reads: " << mapped << endl;
  out << "# Number of twice mapped reads: " << twice_mapped << endl;
  out << "# Number of results: " << results << endl;
	if(only_mapped)
  out << "# Only the reads that were found are listed here (-mapped)." << endl;
  out << "# For each read its name, is printed."<< endl;
  out << "# For each found position of a pattern following information are printed:" << endl;
  out << "# The chromosome and the position in the reference FASTA file, where the read occour, "
			<< "# the start position in the jump index string and a list of the used variations of the vcf file." << endl;
	out << "# [id:altx] denotes that alternative 'x' of the variation 'id' was used."  << endl;
  out << "#READ\tCHROM\tPOS\tJISPOS\tVARIATIONS" << endl;
	
	
	print_row("Time statistics: ",((double)(clock()-time) / CLOCKS_PER_SEC));
  time = clock();
  
	//iterate over results
  for(list<searchresult>::iterator sres = res.match_res.begin(); sres != res.match_res.end(); sres++){   
		unsigned int found_rsc = 0;
		unsigned int rsc = sres->ivl.number_of_results();			
		
		if(rsc == 0) {
			if(!only_mapped)
				out << sres->name << "\t*\t*\t*" << endl;
			continue;		
		}else{
			out << sres->name;
		}

    
    //positions should be unique
    set<int> positions;   
    
    //position from searchintervals
    for(list<searchinterval>::iterator iv = sres->ivl.begin(); iv != sres->ivl.end(); iv++)    
      for(unsigned int i = iv->lr.first; i <= iv->lr.second; i++)
				positions.insert(fmi[i]);
      
    //positions from searchpositions
    for(list<searchposition>::iterator sp = sres->ivl.begin_single(); sp != sres->ivl.end_single(); sp++)
      positions.insert(sp->pos+1);
     
		
    for(set<int>::iterator p = positions.begin(); p!=positions.end(); p++)
      found_rsc +=  forward_search<jia_type>(out,*p, sres->pattern, jis, jia_c, trans, pos_h);  
      
    if(found_rsc != rsc){
      out << "ERROR expectet resultscount=" << rsc << " differ from found resultscount=" << found_rsc << endl; 
    }
  }
  out.close();
	print_row("Mapped reads", mapped);
	print_row("Mapped reads+inv", twice_mapped);
	print_row("Results", results);
	print_row("Time for matching", ((double)(clock()-time) / CLOCKS_PER_SEC));
}




/** \brief checks input of main
 * 
 * arg1 = name of build directory
 * arg2 = name of result file (type matchresult)
 * arg3 = name of output file 
 * 
 * Opt:
 * 
 * \param argc 		argc of main
 * \param argv 		argv of main
 * \param wrongcall	set to true if checking failed
 * \param mapped 	set if -mapped is set
 * 
 * following objects are loaded
 * \param fmi 	fm_index
 * \param jia		jump index array
 * \param res 	matchresults
 * \param trans	transformer class
 * \param jis 	jump index string 
 */
template<class fmi_type, class jia_type>
void checkinput(int & argc,  char** argv, bool & wrongcall, bool & mapped, fmi_type & fmi, jia_type & jia, matchresult & res, transform_id & trans, position_helper & pos_h , string & jis){
  clock_t time 	= clock();

  if(argc <= 3 || argc >= 6){
    cout << "argument count invalid" << endl; 
    wrongcall = true; 
    return;
  }

	wrongcall |= sdsl_load<fmi_type>(argv[1] + FMI_FNAME,fmi);
	wrongcall |= sdsl_load<jia_type>(argv[1] + JIA_FNAME,jia);
	wrongcall |= sdsl_load<matchresult>(argv[2],res);
	wrongcall |= sdsl_load<transform_id>(argv[1] + TRANSFORM_FNAME,trans);
	wrongcall |= sdsl_load<position_helper>(argv[1] + POSITION_FNAME,pos_h);
	wrongcall |= load_string_from_file(argv[1] + JIS_FNAME,jis);
  		
  
  if(argc == 5){
		string arg = argv[4];
    if(arg == "-mapped")
			mapped = true;
		else{
			cout << "parameter: '" + arg +"' unknown.\n";
			wrongcall = true;	
		}
  }  
  print_row("Time for loading fmi,jia,res,transform,jis:",((double)(clock()-time) / CLOCKS_PER_SEC));
	dotline();
}


/**
 * \brief starts forward matcher to analyse results
 * 
 * calls checkinput and startreader
 */
int main(int argc,  char** argv) {
	
	typedef sdsl::csa_wt<sdsl::wt_huff<sdsl::bit_vector,sdsl::bit_vector::rank_1_type,sdsl::select_support_scan<1>,sdsl::select_support_scan<0>>,32,1024*1024> fmi_type;
	typedef compressedjia_rank jia_type;
	
	clock_t time = clock();
  bool wrongcall 	= false;
  bool mapped		= false;
	
	
  fmi_type fmi;  
  jia_type jia;  	 
  matchresult res;
  transform_id trans;
  position_helper pos_h;
	string jis;

  
	checkinput<fmi_type,jia_type>(argc, argv, wrongcall, mapped, fmi, jia, res, trans, pos_h , jis);   

  if(wrongcall){
    cout << "Usage of " << argv[0] << endl <<  "3 parameter needed:" << endl;
    cout << " output dir of construct_index (including " << FMI_FNAME << " " << JIA_FNAME << " " << JIS_FNAME << " " << TRANSFORM_FNAME << " " << POSITION_FNAME <<")" << endl;
		cout << " result file (output of the matcher)" << endl;
		cout << " output file" << endl;
    cout << "Optional param:"<< endl << " -mapped (only reads, that were mapped, will be printed)" << endl;
		cout << "This program prints the positions of the reads in the jump index string" << endl;
    return 0;    
  }  
  
	print_row("Write positions of results in", argv[1]);
	dotline();
	
  startreader<fmi_type,jia_type>(argv[2],argv[3],mapped, fmi, jia, res, trans, pos_h , jis);
	dotline();
	print_row("Time overall: ",((double)(clock()-time) / CLOCKS_PER_SEC));  
	print_row("Output: ", argv[3]);  
}
