/**
	compressedjia_rank.cpp
	Purpose: Compressing JIA with vectors and rank support 

	@author Thomas Büchler
*/
#include "compressedjia_rank.hpp"

using namespace std;
using namespace sdsl;

/**
 * \brief constructs a compressed jia out of a jia 
 * 
 * 
 * singelton[i] indicates if jia.jumptargets has length 1
 * 
 * twin indicates if a  jia.jumptargets has length 2
 * only bits stored for bits that are 0 in singleton
 * twin[i - rank(singelton[i])] for jia.jumptargets[i]
 * 
 * singlejumps, twinjumps, multjumps store the jumptargets
 * 
 * to_var is the same 
 * 
 */
compressedjia_rank::compressedjia_rank(const jumpindexarray & jia){ 
  
  //to_var is the same
  to_var 	= jia.to_var;
  
  
  //number of hashs
  unsigned int size 	= jia.to_var.size();
  //width of the numbers
  unsigned int width 	= jia.jumptargets[0].width();
  

  
  //a bit_vector indicating if a int_vector of jumptargets has size 1
  singleton = bit_vector(size);       
  for(unsigned int i = 0; i < size; i++)
    singleton[i]	= (jia.jumptargets[i].size() == 1);
  
  //init ranksupport
  util::init_support(singlerank,&singleton);
  unsigned int singlecount	= singlerank(size);
	
  
  
  //a bit_vector indicating if a int_vector of jumptargets has size 2
  //only for those indizes stored, which are not singelton
  twin = bit_vector(size-singlecount);  
  for(unsigned int i = 0; i < size; i++)
    twin[i-singlerank(i)]	= (jia.jumptargets[i].size() == 2); 
  
  //init ranksupport
  util::init_support(twinrank,&twin);  
  unsigned int twincount	= twinrank(size-singlecount);
  
  
  //vectors for the jumptargets
  singlejumps	= int_vector<0>(singlecount,0,width);  
  twinjumps 	= int_vector<0>(twincount*2,0,width);  
  multijumps	= vector<int_vector<0>>(size-singlecount-twincount);


  //iterator for the vectors
  unsigned int sing_i	= 0;
  unsigned int twin_i	= 0;
  unsigned int list_i	= 0;
  
  //filling the vectors
  for(unsigned int i = 0; i < size; i++)
    //deciding if i is a single, twin or multijump
    if(singleton[i])      
      singlejumps[sing_i++]	= jia.jumptargets[i][0]; 
    
    else if(twin[i-sing_i]){     
      twinjumps[twin_i++] 	= jia.jumptargets[i][0];
      twinjumps[twin_i++] 	= jia.jumptargets[i][1];
    }    
    else
      multijumps[list_i++] 	= jia.jumptargets[i];    
}

  
/** \brief defaut constructor */
compressedjia_rank::compressedjia_rank(){  
}

/** 
  *\brief returns a vector with the jumptargets
  * 
  * decides if singleton, twin or multi jump
  * 
  * load information from corresponding vector 
  *    
  */
int_vector<0> compressedjia_rank::targets(unsigned int i){  
  unsigned int width = multijumps[0].width();
    //if single return vector with the target
  if(singleton[i])
    return move(int_vector<0>(1,singlejumps[singlerank(i)],width));   
	
  unsigned int twinpos 	= i-singlerank(i);
  unsigned int twinran 	= twinrank(twinpos);
  
  if(twin[twinpos]){
    int_vector<0> out	= int_vector<0> (2,0,width);
    out[0] 	= twinjumps[2*twinran];
    out[1] 	= twinjumps[2*twinran+1];
    return move(out);
  }
  
  //if multi return the vector
  else
    return multijumps[i-singlerank(i)-twinran]; 
};

/** \brief returns if the jump goes to a variation*/
bool compressedjia_rank::to_vari(unsigned int i){
  return to_var[i];
}

/** \brief constructs a jis out of a compressed jis */
jumpindexarray compressedjia_rank::jia(){
  jumpindexarray jia 	= jumpindexarray();
  //to_var is same
  jia.to_var 		= to_var;  
  
  //ask and store all target vectors
  jia.jumptargets 	= vector<int_vector<0>>(to_var.size());   
  for(unsigned int i = 0; i < jia.jumptargets.size(); i++){
    jia.jumptargets[i] 	= targets(i);
  }    
  return move(jia);
}

/** \brief serealizes a jumpindex with sdsls serialize methods */  
compressedjia_rank::size_type compressedjia_rank::serialize(std::ostream& out, sdsl::structure_tree_node* v,std::string name) const {
  
  sdsl::structure_tree_node *child 		= sdsl::structure_tree::add_child(v, name, sdsl::util::class_name(*this));
  compressedjia_rank::size_type written_bytes 	= 0;
  
  written_bytes += singleton	.serialize(out, child, "singleton");
  written_bytes += twin		.serialize(out, child, "twin");
  written_bytes += to_var	.serialize(out, child, "to_var");
  
  written_bytes += singlejumps	.serialize(out, child, "singlejumps");
  written_bytes += twinjumps	.serialize(out, child, "twinjumps"); 
  
  written_bytes += singlerank	.serialize(out, child, "singlerank");
  written_bytes += twinrank	.serialize(out, child, "twinrank"); 
  
  written_bytes += serialize_vector<int_vector<0>>(multijumps,out,child,"multijumps");

  sdsl::structure_tree::add_size(child, written_bytes);
  return written_bytes;
}

/** \brief loads a serialized jumpindex with sdsls methods */
void compressedjia_rank::load(std::istream& in) {
  singleton	.load(in);
  twin		.load(in);
  to_var	.load(in);  
  
  singlejumps	.load(in);
  twinjumps	.load(in);
    
  singlerank	.load(in,&singleton);
  twinrank	.load(in,&twin);  
  
  unsigned int size 		= singleton.size();
  unsigned int list_vec_size 	= size - singlerank(size) - twinrank(twin.size());  
  
  multijumps = vector<int_vector<0>>(list_vec_size);		
  sdsl::load_vector<int_vector<0>>(multijumps,in);  
}

/** \brief for debugging */
ostream& operator <<(ostream& outputStream, const compressedjia_rank& cia){ 
  
  outputStream << "C_JIA [" << endl;
  outputStream << "to_var: " << cia.to_var << endl;  
  outputStream << "single: " << cia.singleton << endl;  
  outputStream << "twin:   " << cia.twin << endl;  
  
  outputStream << "singlejumps: " << cia.singlejumps << endl;
  outputStream << "twinjumps:   " << cia.twinjumps << endl;
  
  
  outputStream << "multijumps:  " << endl;
  for(unsigned int i = 0; i < cia.multijumps.size(); i++){
    outputStream  << cia.multijumps[i].size() << ": " << cia.multijumps[i] << endl;
  }
  outputStream << "]" << endl;

  return outputStream;
}


