/**
	compressedjia_rank.hpp
	Purpose: Compressing JIA with vectors and rank support 

	@author Thomas Büchler
*/
#include <vector>
#include <string>

#include "jumpindexarray.hpp"

#include <sdsl/int_vector.hpp>
#include <sdsl/bit_vectors.hpp>

#ifndef CJI_H
#define CJI_H


/**
 * (In Master Thesis: 6.4.2 Komprimierung mit rank support)
 * \brief compressed jia, compressed infomation that are needed for the jump index 
 * 
 * interfaces:
 * 
 * targets(int i) returns the jumptargets to the # with number i
 * to_vari(int i) returns if the jump goes to a variation
 * 
 * data:
 * 
 * singelton[i] indicates if jia.jumptargets has length 1
 * 
 * twin indicates if a  jia.jumptargets has length 2
 * only bits stored for bits that are 0 in singleton
 * twin[i - rank(singelton[i])] for jia.jumptargets[i]
 * 
 * 
 * to_var indicates if a jump goes to a variation
 * 
 * singlerank, twinrank are ranksupports to the bitvectors
 * 
 * 
 * singlejumps, twinjumps, multjumps store the jumptargets
 * 
 * 
 */
struct compressedjia_rank{
  
  typedef sdsl::int_vector<>::size_type	size_type;
  
  //data  
  
  sdsl::bit_vector		singleton;
  sdsl::bit_vector		twin;
  sdsl::bit_vector		to_var;  
  
  sdsl::int_vector<0>		singlejumps;  
  sdsl::int_vector<0>		twinjumps;  
  std::vector<sdsl::int_vector<0>> 	multijumps;
  
  sdsl::rank_support_v<>	singlerank;
  sdsl::rank_support_v<>	twinrank;
  
  
  
  //interfaces  
  
  /** \brief returns a vector with the jumptargets*/
  sdsl::int_vector<0> 	targets(unsigned int i);  
  /** \brief returns if the jump goes to a variation*/
  bool 			to_vari(unsigned int i); 
  /** \brief uncompresses cia*/
  jumpindexarray 	jia();
    
  
  //constructors  
  /** \brief constructs a compressed jia out of a jia */
  compressedjia_rank(const jumpindexarray & jia);   
  /** \brief defaut constructor */
  compressedjia_rank();
  
  
  //serialization  
  /** \brief serealizes a jumpindex with sdsls serialize methods */  
  size_type 	serialize(std::ostream& out, sdsl::structure_tree_node* v, std::string name) const;		
  /** \brief loads a serialized jumpindex with sdsls methods */
  void 		load(std::istream& in);    
};

std::ostream& operator <<(std::ostream& outputStream, const compressedjia_rank& jia);

#endif
