/**
	compressedjia_select.cpp
	Purpose: Compressing JIA with vectors and select support 

	@author Thomas Büchler
*/
#include "compressedjia_select.hpp"

using namespace std;
using namespace sdsl;

/**
 * \brief constructs a compressed jia out of a jia 
 * 
 * 
 * singelton[i] indicates if jia.jumptargets has length 1
 * 
 * twin indicates if a  jia.jumptargets has length 2
 * only bits stored for bits that are 0 in singleton
 * twin[i - rank(singelton[i])] for jia.jumptargets[i]
 * 
 * singlejumps, twinjumps, multjumps store the jumptargets
 * 
 * to_var is the same 
 * 
 */
compressedjia_select::compressedjia_select(const jumpindexarray & jia){ 
  
  //to_var is the same
  to_var 	= jia.to_var; 
  
  //number jump targets in whole
  int size 	= 0;
  
  for(unsigned int i = 0; i < jia.jumptargets.size(); i++)
    size 	+= jia.jumptargets[i].size();
  
  int width 	= jia.jumptargets[0].width();
  
  // init vectors
  jumps 	= int_vector<0>(size,0,width);
  begin 	= bit_vector(size+1,0);
  begin[size] 	= 1; // safe value at the end
  
  // fill vector
  int jump_i = 0;  
  for(unsigned int i = 0; i < jia.jumptargets.size(); i++){        
    begin[jump_i] 	= 1;     
    for(unsigned int j = 0; j < jia.jumptargets[i].size(); j++)
      jumps[jump_i++] 	= jia.jumptargets[i][j];    
  }  
  
  //init select suppor
  util::init_support<>(beginselect,&begin);  
}
  
/** \brief defaut constructor */
compressedjia_select::compressedjia_select(){  
}


/** 
  *\brief returns a vector with the jumptargets
  * 
  * decides if singleton, twin or multi jump
  * 
  * load information from corresponding vector 
  *    
  */
int_vector<0> compressedjia_select::targets(int i){  
  auto width	= jumps.width();
  
  if(i == 0) //special treatment for i == 0
    return move(int_vector<0>(0,0,width));
 
  //select the position of the ones  
  int j	= beginselect(i);
  int k	= beginselect(i+1);
    
  //init output vector
  int_vector<0> target = int_vector<0>(k-j,0,width);
  //fill it
  for(int l = 0; l + j < k; l++){
    target[l]	= jumps[l+j];
  }  
  return move(target); 
};

/** \brief returns if the jump goes to a variation*/
bool compressedjia_select::to_vari(int i){
  return to_var[i];
}

/** \brief constructs a jia out of a compressed jia */
jumpindexarray compressedjia_select::jia(){
  jumpindexarray jia = jumpindexarray();
  //to_var is same
  jia.to_var 		= to_var;  
  jia.jumptargets 	= vector<int_vector<0>>(to_var.size());   
  
  //ask and store all target vectors
  for(unsigned int i = 0; i < jia.jumptargets.size(); i++){
    jia.jumptargets[i] 	= targets(i);
  }    
  return move(jia);
}

/** \brief serealizes a jumpindex with sdsls serialize methods */
compressedjia_select::size_type compressedjia_select::serialize(std::ostream& out, sdsl::structure_tree_node* v,std::string name) const {
  
  sdsl::structure_tree_node *child 		= sdsl::structure_tree::add_child(v, name, sdsl::util::class_name(*this));
  compressedjia_select::size_type written_bytes = 0;
  
  written_bytes	+= begin	.serialize(out, child, "begin");
  written_bytes	+= to_var	.serialize(out, child, "to_var");
  written_bytes	+= jumps	.serialize(out, child, "jumps");
  written_bytes	+= beginselect	.serialize(out, child, "beginselect");

  sdsl::structure_tree::add_size(child, written_bytes);
  return written_bytes;
}

/** \brief loads a serialized jumpindex with sdsls methods */
void compressedjia_select::load(std::istream& in) {
  begin		.load(in);
  to_var	.load(in);
  jumps		.load(in);
  beginselect	.load(in,&begin);
  
  
}

/** \brief for debugging*/
ostream& operator <<(ostream& outputStream, const compressedjia_select& cia){ 
  
  outputStream << "C2_JIA [" << endl;
  outputStream << "begin: " << cia.begin << endl;  
  outputStream << "jumps:   " << cia.jumps << endl;  
  outputStream << "]" << endl;

  return outputStream;
}


