/**
	conflict_group.cpp
	Purpose: Solfing conflicts with overlapping indels in VCF

	@author Thomas Büchler
*/
#include "conflict_group.hpp"
#include "../methods/iupac_meth.hpp"

#include <iostream>
#include <math.h>       /* pow */
#include <tuple>	/* pair */
#include <deque>	/* pair */

using namespace std;

/** \brief defalut constructor */
conflict_group::conflict_group(){
  vcf_lines 	= deque<vcfl>();
  maxend 	= 0;  
};


/** \brief check if (the next) line is in conflict with the group */
bool conflict_group::conflict(const vcfl & line) const{  
  if(empty())
    return false;  
  if(vcf_lines.front().chrom != line.chrom)
    return false; 
  return   maxend >= line.pos; 
};

  
/** \brief adds a line to the group */
void conflict_group::add(const vcfl & line){  
  unsigned int end 	= line.pos + line.ref.size() -1;	//update maxend
  if(end > maxend)
    maxend 	= end;  
  vcf_lines		.push_back(move(line)); 		//add line  
}; 

/** \brief if group is empty*/
bool conflict_group::empty() const{
  return vcf_lines.empty();
}; 


/**
 * \brief generates the reference string of the conflict group 
 * 
 * reads the ref of the first vcf_line in the queue
 * if chars for whole reference are missing, check next vcf_line
 * 
 * ends if all chars of the reference are read out of the vcf_lines
 */
string conflict_group::generate_ref() const{
  
  unsigned int startpos	= vcf_lines.front().pos;		//reference starts at the pos of the first ref in the queue
  unsigned int refsize 	= maxend - startpos +1;			//reference ends at maxend
    
  
  char ref [refsize];						//buffer for refstring
  int j 	= 0;						// current vcf line
  
  for(unsigned int i = 0; i <  refsize; i++)  			//loop over chars of reference
    
    if(vcf_lines[j].pos + vcf_lines[j].ref.size() -1 >= i + startpos){ 	// what is possible >= what is needed
      ref[i] 	= vcf_lines[j].ref[i + startpos - vcf_lines[j].pos];	// define char
    }else{ // info lies in next vcf line
      j++; // skip to next line
      i--; // redo loop for this char
    } 
    
  return move(string(ref,refsize));  
}


/**
 * \brief generates list of all possible alternative combinations
 * 
 * uses rekursion over vcf lines,
 * 
 * base case is when no vcf line is left.
 * 
 * in each step 2 rekursions can be made
 * 
 * the first is ignoring the line
 * the second is applying the line (only called when possible)
 * 
 * 
 * \param ref		reference string
 * \param p		reference-string-position 	(vcf_lines[0].pos)
 * \param alts		queue where result is stored
 * \param i		line counter 			(rekursion control, default = 0)
 * \param rso		reference-string-offset 	(default = 0)
 * \param tmpalt	temporary alt 			(default = "")
 */
void conflict_group::combine(const string& ref, const unsigned int& p, string& new_alts, unsigned int i, unsigned int off, string tmpalt) const{

  if(i == vcf_lines.size()){ 		//base case, no more vcf lines
    tmpalt	+= ref.substr(off);	//copy end of ref to alt
    new_alts	+= "," +(tmpalt);	//add alt to list
    return;				
  }
  
  //case 1: ignore vcf_line i
    combine(ref, p, new_alts, i+1, off, tmpalt);			// go to rekursion case 1  
  
  //case 2: if vcf_lines[i] is possible apply it
  unsigned int off_l = vcf_lines[i].pos - p;
  if(off <= off_l){ 					// if line i can be applied, apply it
    tmpalt 	+= ref.substr( off, off_l - off);	// copy refstring till begin of ref of this line
    // update reference-string-offset 
		off 	= off_l + vcf_lines[i].ref.size();				// ignore ref of line i
    
			//iterate over alts of this vcfl
		deque<string> altQ = vcf_lines[i].genaltQ();
		for(auto alt = altQ.begin(); alt != altQ.end(); alt++){
			         combine(ref, p, new_alts, i+1, off, tmpalt + *alt);//add alternative and go to rekursion case 2    
		}		 
  }
}

/** \brief this method fueses the alternatives of the snps */
vcfl fuse_snp(deque<vcfl> & vcf_lines){
    
  auto vl = vcf_lines.begin();
  string 	alt	= vl->alt;   
  vl++;
  while(vl != vcf_lines.end()){
    alt += "," + vl->alt;
    vl++;
  }    
  vcf_lines[0].alt = alt;
  return vcf_lines[0];  
}

/** \brief solves conflict, cleans struct this */
vcfl conflict_group::solve(){  
  switch(vcf_lines.size() ){
    case(0):	throw invalid_argument("ERROR: TRY TO SOLVE EMPTY CONFLICT_GROUP");	//empty group cannot be solved			
    case(1):	return(vcf_lines.front());					//just use front
    default:	      

    
    if(only_snps())	return(move(fuse_snp(vcf_lines)));
		
		//generate alt
		//generate altQs
		//for(auto vl = vcf_lines.begin(); vl != vcf_lines.end(); vl++){
		//	vl->genaltQ();
		//}
		
    string 		ref	= generate_ref();
    apply_snps(ref);
    string 	new_alts	= "";   
        combine(ref ,vcf_lines[0].pos,new_alts);
		new_alts = new_alts.substr(new_alts.find (',',1)+1);//ignore first, this is ref  
    
		vcf_lines[0].id += "_extended";
		vcf_lines[0].ref = ref;
		vcf_lines[0].alt = new_alts;
    //return
    return vcf_lines[0]; //make vcfl struct  
  }  
};

/** \brief solves conflict, cleans struct this */
void conflict_group::clear(){
  vcf_lines	.clear();
  maxend 	= 0;  
}

/** \brief extracts vcf_lines that are SNPS and integrates the SNPs in ref.  
 * changes ref and vcf_lines 
 */
void conflict_group::apply_snps(string & ref){
  if(empty()) return;  
  
  for( deque<vcfl>::iterator vl = vcf_lines.begin() ; vl != vcf_lines.end(); vl++)
    if(vl->isSNP()){
      ref[vl->pos - vcf_lines.front().pos] = getSNPLetter(vl->alt,vl->ref[0]);    
      vl = vcf_lines.erase(vl);
      vl--;
    } 
}

/** \brief returns true if all lines are SNPs */
bool conflict_group::only_snps() const{  
  for( auto vl = vcf_lines.begin() ; vl != vcf_lines.end(); vl++)
    if(!vl->isSNP()) return false;    
  return true;
}

/** \brief defiening an string for a variation for streams, for debugging*/
ostream& operator <<(ostream& outputStream, const conflict_group& c){    
  
    outputStream << "/*COGRU:" << endl;
    for(auto l = c.vcf_lines.begin(); l != c.vcf_lines.end(); l++){
      outputStream << "-" << *l;
    }
    outputStream << "mxe: " << c.maxend << endl;
    outputStream << "ref: " << c.generate_ref() << "*/" << endl;
    return outputStream;
}





























