/**
	interval_list.cpp
	Purpose: Struct for "set of intervals" in backward search

	@author Thomas Büchler
*/

#include "interval_list.hpp"

using namespace std;
using namespace sdsl;


/** \brief default constructor*/
interval_list::interval_list(){
  set = list<searchinterval>();
  singleset = list<searchposition>();
};

/** \brief adds a inteval to the set */
void interval_list::add(const searchinterval& iv){
  set.push_back(iv);
}
void interval_list::add(const searchposition& sp){
  singleset.push_back(sp);
}

/** \brief deletes a inteval from the set */
interval_list::iterator interval_list::erase(interval_list::iterator position){
  return set.erase(position);
}
interval_list::iterator_single interval_list::erase(interval_list::iterator_single position){
  return singleset.erase(position);
}


/** \brief merges two sets of intevals*/
void interval_list::merge(interval_list & ivl){
  set.splice(set.end(), ivl.set);
  singleset.splice(singleset.end(), ivl.singleset);
}


/** \brief is the set empty? */
bool interval_list::empty() const{
  return set.empty() && singleset.empty();
}

/** \brief returns iterator to the first element of the set */
interval_list::iterator interval_list::begin(){
  return set.begin();
}
interval_list::iterator_single interval_list::begin_single(){
  return singleset.begin();
}


/** \brief returns iterator to the last+1 element of the set */
interval_list::iterator interval_list::end(){
  return set.end();
}
interval_list::iterator_single interval_list::end_single(){
  return singleset.end();
}


/** \brief returns constant iterator to the first element of the set */
interval_list::const_iterator interval_list::begin() const{
  return set.begin();
}
interval_list::const_iterator_single interval_list::begin_single() const{
  return singleset.begin();
}


/** \brief returns constant iterator to the last+1 element of the set */
interval_list::const_iterator interval_list::end() const{
  return set.end();
}
interval_list::const_iterator_single interval_list::end_single() const{
  return singleset.end();
}

/** \brief calculates number of mapping results of this set */
unsigned int interval_list::number_of_results(){
  
  unsigned int result_count = 0;
  for(auto iv = begin(); iv != end(); iv++) 
    result_count	+=	iv->lr.second - iv->lr.first + 1;
  
  result_count += singleset.size();
  
  return result_count;
}

/** \brief serialize a variation*/
interval_list::size_type interval_list::serialize(std::ostream& out, sdsl::structure_tree_node* v,std::string name) const {
  
  sdsl::structure_tree_node *child = sdsl::structure_tree::add_child(v, name, sdsl::util::class_name(*this));
  size_type written_bytes = 0;
  
  
  int_vector<32> setsizes = int_vector<32>(2,0,32);
  setsizes[0] = set.size();
  setsizes[1] = singleset.size();
  
  written_bytes += setsizes.serialize(out, child, "setsizes");
  
  for(auto iv = begin(); iv != end(); ++iv)
    written_bytes += iv->serialize(out,child,"iv");

  for(auto sp = begin_single(); sp != end_single(); ++sp)
    written_bytes += sp->serialize(out,child,"sp");
  
  sdsl::structure_tree::add_size(child, written_bytes);
  return written_bytes;
}

/** \brief laod a serialized variation*/
void interval_list::load(std::istream& in) {
  int_vector<32> setsizes;
  setsizes.load(in);	
  
  set = list<searchinterval>();
  singleset = list<searchposition>();
  for(unsigned int i = 0; i < setsizes[0]; i++){
    searchinterval iv = searchinterval(searchinterval::bounds(0,0));
    iv.load(in);
    add(iv);    
  }
  
  for(unsigned int i = 0; i < setsizes[1]; i++){
    searchposition sp = searchposition(0);
    sp.load(in);
    add(sp);    
  }  
}

/** \brief generates a string with the intervals, for depugging */
ostream& operator <<(ostream& outputStream, const interval_list& ivl){ 
  outputStream << "Interval_list:" << endl;
  for(auto i = ivl.begin(); i != ivl.end(); i++){  
		outputStream << "-bwt_iv: [" << i->lr.first << ", " << i->lr.second<< "]" << endl;
	}
	for(auto i = ivl.begin_single(); i != ivl.end_single(); i++){  
		outputStream << "-position: " << i->pos +2 << " ret: " ;
		
		for(auto j = i->ret.begin(); j!= i->ret.end(); j++)
			outputStream << *j + 2 << " ";
		outputStream << endl;
	}
  return outputStream;
}
