/**
	jumpindexarray.cpp
	Purpose: Struct for jump-index-array

	@author Thomas Büchler
*/

#include "jumpindexarray.hpp"

using namespace std;
using namespace sdsl;



  
/** \brief defaut constructor */
jumpindexarray::jumpindexarray(){
  jumptargets 	= vector<int_vector<0>>(0);
  to_var 	= bit_vector(0);
}

/** \brief copy constructor */
jumpindexarray::jumpindexarray(const jumpindexarray & jia){
  jumptargets 	= jia.jumptargets;
  to_var 	= jia.to_var;
};

jumpindexarray::jumpindexarray(vector<int_vector<0>> jumptargets, bit_vector to_var):jumptargets(jumptargets), to_var(to_var){};


/** \brief returns a vector with the jumptargets*/
int_vector<0> jumpindexarray::targets(unsigned int i){
  return jumptargets[i];
};


/** \brief returns if the jump goes to a variation*/
bool jumpindexarray::to_vari(unsigned int i){
  return to_var[i];

}

jumpindexarray jumpindexarray::get_transformed_jia(int_vector<0> & permutation){
	if(permutation.size() < jumptargets.size() )
		throw invalid_argument( "ERROR: jia can not be transformed. Given permutation is invalid.");  
	
	vector<int_vector<0>> trans_jumptargets		= vector<int_vector<0>>(jumptargets.size());
  bit_vector trans_to_var 									= bit_vector(to_var.size());
	
	//transform to_var
	for(unsigned int i = 0; i < to_var.size(); i++)
		trans_to_var[permutation[i]] = to_var[i];
	
	//transform jumptargets
	for(unsigned int i = 0; i < jumptargets.size(); i++){
		int_vector<0> trans_target = int_vector<0>(jumptargets[i].size(),0,jumptargets[i].width());
		
		for(unsigned int j = 0; j < jumptargets[i].size(); j++)
			trans_target[j] =  permutation[jumptargets[i][j]];
		
		trans_jumptargets[permutation[i]] = trans_target;
	}
	
	return jumpindexarray(move(trans_jumptargets),move(trans_to_var));
}

jumpindexarray jumpindexarray::get_inverted_jia(){
	vector<int_vector<0>> inv_jumptargets		= vector<int_vector<0>>(jumptargets.size());	
	
	//count how often a hash is targeted
	int_vector<32> targeted_count = int_vector<32>(jumptargets.size(),0,32);
	for(unsigned int i = 0; i < jumptargets.size(); i++)
		for(unsigned int j = 0; j < jumptargets[i].size(); j++)
			targeted_count[ jumptargets[i][j] ]++;
	
	//initialize new target vectors
	for(unsigned int i = 0; i < inv_jumptargets.size(); i++)
		inv_jumptargets[i] = int_vector<0>(targeted_count[i],0,jumptargets[0].width());
	
	for(unsigned int i = 0; i < jumptargets.size(); i++)
		for(unsigned int j = 0; j < jumptargets[i].size(); j++){
			unsigned int k = jumptargets[i][j];
			inv_jumptargets[k][inv_jumptargets[k].size()-(targeted_count[k]--)] = i;			
		}
	
	return jumpindexarray(move(inv_jumptargets),to_var);
}





  /** \brief serealizes a jumpindex with sdsls serialize methods */
jumpindexarray::size_type jumpindexarray::serialize(std::ostream& out, sdsl::structure_tree_node* v,std::string name) const {
  
  sdsl::structure_tree_node *child = sdsl::structure_tree::add_child(v, name, sdsl::util::class_name(*this));
  jumpindexarray::size_type written_bytes = 0;


  written_bytes 	+= to_var.serialize(out, child, "to_var");	
  written_bytes 	+= serialize_vector<int_vector<0>>(jumptargets,out,child,"jumptargets");

  sdsl::structure_tree::add_size(child, written_bytes);
  return written_bytes;
}


/** \brief loads a serialized jumpindex with sdsls methods */
void jumpindexarray::load(std::istream& in) {
  to_var	.load(in);
  jumptargets 	= vector<int_vector<0>>(to_var.size());		
  sdsl		::load_vector<int_vector<0>>(jumptargets,in);		
}


/** \brief generates a string with the targets, for depugging */
ostream& operator <<(ostream& outputStream, const jumpindexarray& jia){ 
  outputStream << "JIA (to_var:jumptargets):" << endl;  
  for(unsigned int i = 0; i < jia.jumptargets.size(); i++){
    outputStream << i << " " << jia.to_var[i] << ": " << jia.jumptargets[i] << endl;
  }
  return outputStream;
}

/** \brief cleans storage */
void jumpindexarray::clear(){  
  vector<sdsl::int_vector<0>> foo 	= vector<int_vector<0>>(0);
  jumptargets				.swap(foo);
  sdsl::util				::clear(to_var);
}


