/**
	kmer_index.cpp
	Purpose: Struct for kmer index

	@author Thomas Büchler
*/

#include "kmer_index.hpp"

#include <stdexcept>	// invalid_argument
#include <list>
#include <math.h>       // pow

#include "../structs/searchinterval.hpp"

#include <sdsl/int_vector.hpp>

using namespace std;
using namespace sdsl;

/** \brief default constructor*/
kmer_index::kmer_index(){}

/** \brief default constructor*/
kmer_index::kmer_index(int k): k(k){
	unsigned int kmer_count = ((unsigned int)round(pow(kmer_index::letter_count,k)));    
	kmer = vector<interval_list>(kmer_count);	
}

/** \brief copy constructor*/
kmer_index::kmer_index(int k, vector< interval_list> & ivl): k(k), kmer(ivl){
  if(((unsigned int)round(pow(letter_count,k))) != ((unsigned int) ivl.size()))
   throw invalid_argument("readlength and prec do not match (kmer_index::kmer_index)");
}
/** \brief copy constructor*/
kmer_index::kmer_index(int k, vector< interval_list> ivl): k(k), kmer(ivl){
  if(((unsigned int)round(pow(letter_count,k))) != ((unsigned int) ivl.size()))
   throw invalid_argument("readlength and prec do not match (kmer_index::kmer_index)");
}

/** \brief copy constructor
kmer_index::kmer_index(int k, vector< interval_list>* ivl): k(k){
  if(((unsigned int)round(pow(letter_count,k))) != ((unsigned int) ivl->size()))
   throw invalid_argument("readlength and prec do not match (kmer_index::kmer_index)");
	kmer = *ivl;	
}*/


/** \brief returns the number of a base*/
unsigned int kmer_index::base2int(const char& c) {
  switch(c){
    case 'A': return 0;
    case 'C': return 1;
    case 'G': return 2;
    case 'T': return 3;
    //case 'N': return 4;
  }
  throw invalid_argument("Unallowed Letter in (kmer_index::base2int): >" + string(1,c) + "<");
}

/** \brief returns a base of given number*/
char kmer_index::int2base(const unsigned int i) {
  switch(i){
    case 0: return 'A';
    case 1: return 'C';
    case 2: return 'G';
    case 3: return 'T';
    //case 4: return 'N';
  }
  throw invalid_argument("Unallowed number (kmer_index::int2base)");
}

/** \brief returns the number of a basestring
 * AAAA = 0
 * CAAA = 1
 * GAAA = 2
 * ...
 */
unsigned int kmer_index::get_count(const string& s){  
  unsigned int num = 0;  
  for(unsigned int i = s.length(); i > 0; i--){
    num *= letter_count;
    num += base2int(s[i-1]);
  }  
  return num;
}

/** \brief returns the basestring of given number
 * eg. k = 4
 * 0 = AAAA
 * 1 = CAAA
 * ...
 */
string kmer_index::get_string(unsigned int count, const unsigned int k){
  string str = string(k,'A');  
  for(unsigned int i = 0 ; i < k ; i++){
    str[i] = int2base(count % letter_count);
    count /= letter_count;
  }
  return move(str);
}

/** \brief returns the interval_list for a  given basestring*/
interval_list kmer_index::get_ivlist(const string& s) const{
  if(k != s.length())
    throw invalid_argument("readlength and basestring do not match (kmer_index::get_ivlist)");
  
  return kmer[get_count(s)];
}

/** \brief removes each entry that mappes 'N' more then n times */
void kmer_index::remove_n(const int & n){
  for( vector<interval_list>::iterator ivl = kmer.begin(); ivl != kmer.end(); ivl++){
    
    for(list<searchinterval>::iterator iv = ivl->begin(); iv != ivl->end(); iv++ )
      if(iv->N > n){
				iv = ivl->set.erase(iv);
				iv--;
      }
    
    for(list<searchposition>::iterator sp = ivl->begin_single(); sp != ivl->end_single(); sp++ )
      if(sp->N > n){
				sp = ivl->singleset.erase(sp);
				sp--;
      }
  }
}



/** \brief serialize struct*/
kmer_index::size_type kmer_index::serialize(ostream& out, structure_tree_node* v, string name) const{
  structure_tree_node *child = structure_tree::add_child(v, name, sdsl::util::class_name(*this));
  size_type written_bytes = 0;
  

  int_vector<32> k_vec = int_vector<32>(1,k,32);
    
  written_bytes += k_vec.serialize(out, child, "k");
  written_bytes += serialize_vector<interval_list>(kmer,out,child,"kmer_index");

  structure_tree::add_size(child, written_bytes);
  return written_bytes;
}

/** \brief load serialized struct*/
void kmer_index::load(istream& in){
  int_vector<32> k_vec;
  k_vec.load(in);  
  k = k_vec[0];  
  
  kmer = vector<interval_list>(pow(letter_count,k));		
  load_vector<interval_list>(kmer,in);	
}
