/**
	position_helper.cpp
	Purpose: Linking jis posions to fasta posiontions, chromosoms and variations

	@author Thomas Büchler
*/

#include "position_helper.hpp"

using namespace std;
using namespace sdsl;

/**
 * \brief converts construct_position_helper to position_helper
 * 
 * concatenates 
 * deque<int> fa_pos_jis,fa_pos_tmp1,fa_pos_tmp2 
 * to 
 * sdsl::int_vector fa_pos
 * 
 * converts 
 * deque<string> chromosome_names
 * to
 * vector<string> chromosome_names
 * 
 * converts 
 * deque<int> seq_seperator
 * to
 * sdsl::int_vector seq_seperator
 * 
 * concatenates 
 * deque<int> vcf_id_jis,vcf_id_tmp1
 * to 
 * sdsl::int_vector vcf_id
 * 
 * concatenates
 */
position_helper construct_position_helper::convert(){
	position_helper ph = position_helper();	
	
	//concatenates fa_pos
	unsigned int size = fa_pos_jis.size() + fa_pos_tmp1.size() + fa_pos_tmp2.size(); 	
	ph.fa_pos = int_vector<0>(size,0,32);	//TODO may not just 32
	for(unsigned int i = 0; i < fa_pos_jis.size(); i++)
		ph.fa_pos[i] = fa_pos_jis[i];	
	for(unsigned int i = 0; i < fa_pos_tmp1.size(); i++)
		ph.fa_pos[i + fa_pos_jis.size()] = fa_pos_tmp1[i];
	for(unsigned int i = 0; i < fa_pos_tmp2.size(); i++)
		ph.fa_pos[i + fa_pos_jis.size()+ fa_pos_tmp1.size()] = fa_pos_tmp2[i];
	
	//convert chromosome_names
	ph.chromosome_names = vector<string>(chromosome_names.size());
	for(unsigned int i = 0; i < chromosome_names.size(); i++)
		ph.chromosome_names[i] = chromosome_names[i];
	
	//convert seq_seperator
	ph.seq_seperator =  int_vector<0>(seq_seperator.size(),0,32);	
	for(unsigned int i = 0; i < seq_seperator.size(); i++)
		ph.seq_seperator[i] = seq_seperator[i];
		
	//concatenates vcf_id
	ph.vcf_id = vector<string>(vcf_id_jis.size() + vcf_id_tmp1.size());
	for(unsigned int i = 0; i < vcf_id_jis.size(); i++)
		ph.vcf_id[i] = vcf_id_jis[i];
	for(unsigned int i = 0; i < vcf_id_tmp1.size(); i++)
		ph.vcf_id[i+vcf_id_jis.size()] = vcf_id_tmp1[i];
	
	return move(ph);
}

/**
 * \brief converts a string to a int_vector<8>
 */
int_vector<8> s2iv(const string & s){
	int_vector<8> iv = int_vector<8>(s.size(),0,8);
	for(unsigned int i = 0; i < s.size(); i++)
		iv[i] = s[i];
	return move(iv);
}

/**
 * \brief converts a int_vector<8> to a string
 */
string iv2s(const int_vector<8> & iv){
	string s;
	for(unsigned int i = 0; i < iv.size(); i++)
		s += iv[i];
	return move(s);
}

/**
 * \brief converts a vector<int_vector<8>> to a vector<string>
 */
vector<string> ivv2sv(const vector<int_vector<8>> & ivv){
	vector<string> sv = vector<string>(ivv.size());
	for(unsigned int i = 0; i < ivv.size(); i++)
		sv[i] = iv2s(ivv[i]);
	return move(sv);
}

/**
 * \brief converts a vector<string> to a vector<int_vector<8>>
 */
vector<int_vector<8>> sv2ivv(const vector<string> & sv){
	vector<int_vector<8>> ivv = vector<int_vector<8>>(sv.size());
	for(unsigned int i = 0; i < sv.size(); i++)
		ivv[i] = s2iv(sv[i]);
	return move(ivv);
}

/**
 * \brief serializes position_helper with help of sdsl
 */
position_helper::size_type position_helper::serialize(std::ostream& out, structure_tree_node* v, std::string name) const{
	structure_tree_node *child = structure_tree::add_child(v, name, util::class_name(*this));
  position_helper::size_type written_bytes = 0;
  
  written_bytes += fa_pos.serialize(out, child, "fa_pos");
  written_bytes += seq_seperator.serialize(out, child, "seq_seperator");
	
	
  int_vector<64> counts = int_vector<64>(2,0,64);
	counts[0] = vcf_id.size();
	counts[1] = chromosome_names.size();
  written_bytes += counts.serialize(out, child, "counts");
	
	vector<int_vector<8>> ivv = sv2ivv(vcf_id);
  written_bytes += serialize_vector<int_vector<8>>(ivv,out,child,"vcf_id");
	ivv = sv2ivv(chromosome_names);
  written_bytes += serialize_vector<int_vector<8>>(ivv,out,child,"chromosome_names");
  

  structure_tree::add_size(child, written_bytes);
  return written_bytes;
}

/**
 * \brief loads position_helper with help of sdsl
 */
void position_helper::load(std::istream& in){
	fa_pos.load(in);
	seq_seperator.load(in);
	
	int_vector<64> counts = int_vector<64>(2,0,64);
	counts.load(in);
	
	vector<int_vector<8>> ivv = vector<int_vector<8>>(counts[0]);		
  load_vector<int_vector<8>>(ivv,in);	
	vcf_id = ivv2sv(ivv);
	
	ivv = vector<int_vector<8>>(counts[1]);		
  load_vector<int_vector<8>>(ivv,in);	
	chromosome_names = ivv2sv(ivv);	
}

/**
 * \brief prints position_helper
 * for debugging
 */
std::ostream& operator <<(std::ostream& out, const position_helper& ph){
	for( unsigned int i = 0; i < ph.fa_pos.size(); i++){
		out << i << '\t' << ph.fa_pos[i];
		if(i < ph.vcf_id.size())
			out << '\t'<< ph.vcf_id[i];
		out << endl;		
	}
	
	for( unsigned int i = 0; i < ph.chromosome_names.size(); i++){
		out << ph.chromosome_names[i] << endl;		
	}
	out << ph.seq_seperator << endl;
	return out;
}


