/**
	sequence_streams.cpp
	Purpose: Storing information about the streams to vcf-files, jis-file and tmp-files

	@author Thomas Büchler
*/

#include "sequence_streams.hpp"
#include "../programs/filenames.hpp"
#include "../methods/iupac_meth.hpp" //matches

#include <iostream>	//cout TODO weg

using namespace std;

sequence_streams::sequence_streams(){
}


sequence_streams::sequence_streams(std::string& fa_filename, std::string& vcf_dirname, std::string out_dirname): fa_filename(fa_filename){

  //init filenames  
	snp_filename	= vcf_dirname + SNP_FNAME;
	ind_filename	= vcf_dirname + IND_FNAME;
	sv_filename		= vcf_dirname + SV_FNAME;
	
  jis_filename			= out_dirname + JIS_FNAME;
	tmp_lay1_filename	=	out_dirname + "/.tmp_lay1";
	tmp_lay2_filename	= out_dirname + "/.tmp_lay2";
	

	
	//init counters
	jis_markers		= 0;
	lay1_markers	= 0;
	lay2_markers	= 0;	
	current_fa_pos	= 1;
	fa_chrom			= "NO_CHROM_REACHED";
	
	write_in_jis = true;
	   write_in_lay1_ended = false;
}

void sequence_streams::open_for_jis(){
	//open ifstreams
	fa_in		.open(fa_filename,ios_base::in);   
  snp_in	.open(snp_filename,ios_base::in);
  ind_in	.open(ind_filename,ios_base::in);
  sv_in		.open(sv_filename,ios_base::in);
	
	//read first chrom name
	copy_till_next_chrom();
	
	//open ofstreams
	jis_out				.open(jis_filename,ios_base::out);	
  tmp_lay1_out	.open(tmp_lay1_filename,ios_base::out);	
  tmp_lay2_out	.open(tmp_lay2_filename,ios_base::out);
}

void sequence_streams::open_for_prepvcf(string & vcf_filename){
	
	//open ofstreams  
	snp_out	.open(snp_filename,ios_base::out);
  ind_out	.open(ind_filename,ios_base::out);
  sv_out	.open(sv_filename,ios_base::out);
	
	snp_out 	<< "##SNPs of file: " << vcf_filename<< endl;
	snp_out 	<< "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n";
	ind_out 	<< "##Indels of file: " << vcf_filename<< endl;
	ind_out		<< "##Overlapping INDELS were combined to one INDEL that is containing all combinations" << endl;
	ind_out 	<< "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n";
	sv_out 		<< "##Structural variants of file: " << vcf_filename<< endl;
	sv_out 		<< "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n";
}

void sequence_streams::close(){
	//input streams
  fa_in.close();   
  snp_in.close();
  ind_in.close();
  sv_in.close();
  tmp_lay1_in.close();
  tmp_lay2_in.close();
	
	
	//output streams
  jis_out.close();
  tmp_lay1_out.close();
  tmp_lay2_out.close();	
	snp_out	.close();
  ind_out	.close();
	sv_out	.close(); 
}

/** \brief writes char to jis */
void sequence_streams::write_char(char c){ 
  //write in output_s_hash_to_v
	if(write_in_jis)
		jis_out   << c;  
	else 
		tmp_lay1_out << c;
	check_lay1_ended();
}

/** \brief writes marker jis */
void sequence_streams::write_marker(){ 
  //write in output_s_hash_to_v
	if(write_in_jis){
		jis_out   << '#';  
		jis_markers++;
	}
	else {
		tmp_lay1_out << '#';
		lay1_markers++;
	}	
}

/** \brief writes an alternative to tmp_lay2_out */
void sequence_streams::write_alternative(string & alt){
	tmp_lay2_out 	<< alt;     
	tmp_lay2_out	<< '#';
	lay2_markers++;
}

void sequence_streams::check_lay1_ended(){
	if(!write_in_jis && current_fa_pos >= write_in_lay1_end){
		tmp_lay1_out	<< '#';
		lay1_markers++;
		write_in_jis = true;
		write_in_lay1_ended = true;
		write_in_lay1_end = current_fa_pos;
	}
}

/** \brief copy the chars in fa file to jis till pos is reached*/
void sequence_streams::copy_till_pos(unsigned int pos){
	char c;  
  while(current_fa_pos < pos){
	  
		if(!fa_in.get(c)) throw invalid_argument( "ERROR\tFA STREAM ENDED BEVOR POSITION " + to_string(pos) +" WAS REACHED");      
		if (c == '\n') 	 continue;      
		else
		{
			try
			{
				iupac_number(c);
				write_char(c);
				current_fa_pos++;
			}catch(invalid_argument &ia){
				throw invalid_argument("ERROR\tUNEXPECTED SYMBOL: '" + to_string(c) + "' IN FA" + " at pos " +to_string(current_fa_pos) + " in ctp");
			}
		}
		
		check_lay1_ended();

  }
}

/** \brief copy the chars in fa file to jis till end is reached
void sequence_streams::copy_till_end(){
	char c;  
  while(fa_in.get(c)){
		
		if(!write_in_jis && current_fa_pos == write_in_lay1_end){
			tmp_lay1_out	<< '#';
			lay1_markers++;
			write_in_jis = true;
			         write_in_lay1_ended = true;
		}		
		
    if (c == '\n') 	 
			continue;  
		
    else if (c == 'A' || c == 'T' || c == 'G'|| c == 'C'|| c == 'N'){
      write_char(c);        
      current_fa_pos++;    
			
    } else if(c == '>'){
      string chrom;
      getline(fa_in, chrom);
      
    }else{      
      throw invalid_argument("ERROR\tUNEXPECTED SYMBOL: '" + to_string(c) + "' IN FA" + " at pos " +to_string(current_fa_pos) + " in cte");      
    }
  }
}*/

/** \brief skippes to the next chrom (maked by '>') in the .fa file and copies the bases to the .jis file
 *  
 * \return name of the next chrom in the fa file or "END_OF_FILE", if no more chrom
 */
void sequence_streams::copy_till_next_chrom(){
  char c; 
  while(fa_in.get(c)){
		
		
		
    
    if(c == '>'){
      //read gene name
      getline(fa_in, fa_chrom);
			fa_chrom = fa_chrom.substr(0,fa_chrom.find('|')); //cut additional infos
      last_chrom_length = current_fa_pos;
			current_fa_pos = 1;  
      return;      
    } 
    else if (c == '\n')	continue;
    else if (c == 'A' || c == 'T' || c == 'G'|| c == 'C'|| c == 'N'){
			write_char(c);   
      current_fa_pos++;   
		}
    else		throw invalid_argument("ERROR\tUNEXPECTED SYMBOL: '" + string(1,c) + "' IN FA" + " at pos " +to_string(current_fa_pos) + " in ctnc");      
    
	check_lay1_ended();
}
  //chrom not found
  fa_chrom = "END_OF_FILE";
	last_chrom_length = current_fa_pos;
	current_fa_pos = 1;  
}

bool sequence_streams::check_ref(char& c){
	current_fa_pos++;     
	char ref; 
	
	//read ref
	for( unsigned int i = 0; i < 1; i++){      
		if(!fa_in.get(ref))	throw invalid_argument("ERROR\tFA STREAM ENDED BEVOR REFERENCE WAS READ");         
		if(ref == '\n'){
			i--;
			continue;   	
		}            
	} 	
	//check if ref matches c
	if(!matches(ref,c)){
		//write ref in jis
		write_char(ref);
		return false;
	}
	return true;
}

bool sequence_streams::check_ref(string & s){
	current_fa_pos	+= s.size();   
	
	string ref;
  char c;  
	bool ok = true;
   
	//read ref
	for( unsigned int i = 0; i < s.size(); i++){      
		if(!fa_in.get(c))	throw invalid_argument("ERROR\tFA STREAM ENDED BEVOR REFERENCE WAS READ");         
		if(c == '\n'){
			i--;
			continue;   	
		}          
		ref += c;   
	}      
	
	//check if ref matches s
	for( unsigned int i = 0; i < ref.size(); i++){
		if(!matches(ref[i],s[i])){
			ok = false; 
			break;
		}
	}
	
	if(!ok){
		//write ref in jis
		for( unsigned int i = 0; i < ref.size(); i++)
			write_char(ref[i]);
	}return ok;
}




