/**
	transform.hpp
	Purpose: Store information to easily convert diffrent numberings of the markers

	@author Thomas Büchler
*/

#include "transform.hpp"


using namespace std;
using namespace sdsl;

transform_id::transform_id(){}

/** \brief calc all vectors out ouf rb2ps */
void transform_id::make_vectors(){
  ps2rb = make_ps2rb();
  rs2rb = make_rs2rb();
  rb2rs = invertVector(rs2rb);
  rs2ps = make_rs2ps();
}

/** \brief invert rb2ps to ps2rb*/
map<unsigned int,unsigned int> transform_id::make_ps2rb(){
  map<unsigned int,unsigned int> out = map<unsigned int,unsigned int>();  
  for(unsigned int i = 0; i < rb2ps.size(); i++)
    out.insert(pair<unsigned int,unsigned int>(rb2ps[i],i));
  return move(out);
}

/**
 *\brief calc rs2rb by using ps2rb
 * 
 * run over map elements (sorted by hash_seq_pos)
 * you get the rank(#,bwt) in order of rank(#,sequence)
 */
int_vector<0> transform_id::make_rs2rb(){  
  unsigned int numberOfHash = rb2ps.size();  
  int_vector<0> rs2rb = int_vector<0>(numberOfHash,0,bitsneeded(numberOfHash));
  
  unsigned int i = 0;
  for( auto iter = ps2rb.begin(); iter != ps2rb.end(); iter++ )
    rs2rb[i++] = iter->second; 
  
  return move(rs2rb);     
};


/**
 *\brief calc rs2ps by using ps2rb
 * 
 * run over map elements (sorted by hash_seq_pos)
 */
int_vector<0> transform_id::make_rs2ps(){  
  unsigned int numberOfHash = rb2ps.size();  
  int_vector<0> rs2ps = int_vector<0>(numberOfHash,0,bitsneeded(seqsize));
  
  unsigned int i = 0;
  for( auto iter = ps2rb.begin(); iter != ps2rb.end(); iter++ )
    rs2ps[i++] = iter->first; 
  
  return move(rs2ps); 
}

unsigned int transform_id::ps2rs(unsigned int ps){
	return rb2rs[ps2rb.find(ps)->second];
}



/**
 * \brief finds rank of the next marker to position pos with binary search
 */
unsigned int transform_id::next_marker(unsigned int pos){
	//left and rigth bound
	unsigned int lb = 0;
	unsigned int rb = rs2ps.size()-1;
	
	while(lb != rb){
		//middle
		unsigned int m = (lb+rb)/2; 
		//jis pos of middle
		unsigned int mpos = rs2ps[m];
		
		if(mpos < pos){
			lb = m + 1;
		}else if(mpos > pos){
			rb = m;
		}else{
			return m; 
		}		
	}
	return lb; 	
};




/** \brief serialize */
transform_id::size_type transform_id::serialize(ostream& out, structure_tree_node* v, string name) const{
  
  structure_tree_node *child = structure_tree::add_child(v, name, util::class_name(*this));
  transform_id::size_type written_bytes = 0;
  
  written_bytes += rb2ps .serialize(out, child, "rb2ps");
  
  int_vector<32> seqsize_v = int_vector<32>(1,seqsize,32);
  written_bytes += seqsize_v .serialize(out, child, "seqsize_v");
  
  structure_tree::add_size(child, written_bytes);
  return written_bytes;
}


/** \brief load */
void transform_id::load(istream& in){  
  rb2ps.load(in);  
  
  int_vector<32> seqsize_v;
  seqsize_v.load(in);  
  seqsize = seqsize_v[0];
  
  make_vectors();  
}











